# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

"""
`chaosmagpy.model_utils` provides functions for analyzing and estimating
geomagnetic field models in general, and the CHAOS geomagnetic field model in
particular.

.. autosummary::
    :toctree: functions

    design_matrix
    design_gauss
    colloc_matrix
    augment_breaks
    pp_from_bspline
    synth_from_pp
    synth_values
    legendre_poly
    power_spectrum
    degree_correlation
    sensitivity

"""

import numpy as np
import warnings
from scipy.interpolate import BSpline, PPoly
from . import config_utils


def design_matrix(knots, order, n_tdep, time, radius, theta, phi,
                  n_static=None, source=None):
    """
    Returns matrices that connect radial, colatitude and azimuthal field
    components on a grid with `radius`, `theta` (colatitude) and `phi`
    to the spherical harmonics expansion of a potential.

    The potential is time-dependent on large length-scales (`n` <= `n_tdep`)
    and static on small length-scales (`n_tdep` < `n` <= `n_static`). The
    time-dependent part uses a B-spline representation of order `k`.

    Parameters
    ----------
    knots : ndarray, shape >= (k+1,)
        Knot vector with adequate endpoint multiplicity.
    order : int, positive
        Order `k` of B-spline basis functions (4 = cubic).
    n_tdep : int, positive
        Maximum degree of the time-dependent field.
    time : ndarray, shape (N,)
        Time vector of the `N` data points in days.
    radius : ndarray, shape (N,)
        Radius vector of the `N` data points in kilometers.
    theta : ndarray, shape (N,)
        Data colatitude vector in degrees :math:`[0^\\circ,180^\\circ]`.
    phi : ndarray, shape (N,)
        Data longitude vector in degrees.
    n_static : int, positive, optional
        Maximum degree of static field (default is ``None``,
        `n_static` > `n_tdep`).
    source : {'internal', 'external'}, optional
        Magnetic field source (default is internal).

    Returns
    -------
    G_radius, G_theta, G_phi : ndarray, shape (N, ...)
        Forward matrix for :math:`B_r`, :math:`B_{\\theta}` and
        :math:`B_{\\phi}`.

    Raises
    ------
    ValueError
        If degree of static field is smaller than the degree of the time-dep.
        field. In order to exclude the static field and obtain a purely
        time-dependent potential, use optional argument ``n_static=None`` or
        leave out completely.

    See Also
    --------
    design_gauss

    """

    assert time.shape == radius.shape == theta.shape == phi.shape
    assert time.ndim == 1
    assert np.amin(theta) > 0. and np.amax(theta) < 180.

    n_data = time.size  # number of data points

    if n_static:
        if n_static < n_tdep:
            raise ValueError("Degree of static field must be greater than "
                             "degree of time-dependent field.")
        elif n_static == n_tdep:
            warnings.warn("Static and time-dependent field are of same "
                          "degree. Ignoring static field.")

    if n_static is None:
        n_static = n_tdep

    if source is None:
        source = 'internal'

    n_coeff_tdep = n_tdep * (n_tdep + 2)  # number of t-dep field coefficients
    n_coeff_static = n_static * (n_static + 2)  # number of static field coeff

    # compute matrices that connect harmonics expansion with field values
    A_radius, A_theta, A_phi = design_gauss(radius, theta, phi,
                                            nmax=n_static, source=source)

    # compute collocation matrix of B-spline segments
    collmat = colloc_matrix(time, knots, order)
    n_segment = collmat.shape[1]  # number of B-spline basis functions

    # allocate memory for output matrices
    G_radius = np.empty((n_data, (n_segment - 1)*n_coeff_tdep
                        + n_coeff_static))
    G_theta = np.empty((n_data, (n_segment - 1)*n_coeff_tdep + n_coeff_static))
    G_phi = np.empty((n_data, (n_segment - 1)*n_coeff_tdep + n_coeff_static))

    n_col = 0  # counter for columns
    for coeff in range(n_coeff_tdep):
        # for each time-dependent expansion coefficient, multiply with
        # collocation matrix use broadcasting to match shapes
        G_radius[:, n_col:n_col+n_segment] = A_radius[:, coeff, None] * collmat
        G_theta[:, n_col:n_col+n_segment] = A_theta[:, coeff, None] * collmat
        G_phi[:, n_col:n_col+n_segment] = A_phi[:, coeff, None] * collmat

        n_col += n_segment

    # add static background field
    G_radius[:, n_col:] = A_radius[:, n_coeff_tdep:n_coeff_static]
    G_theta[:, n_col:] = A_theta[:, n_coeff_tdep:n_coeff_static]
    G_phi[:, n_col:] = A_phi[:, n_coeff_tdep:n_coeff_static]

    return G_radius, G_theta, G_phi


def colloc_matrix(x, knots, order, deriv=None):
    """
    Create collocation matrix of a univariate function on `x` in terms  of a
    B-spline representation of order `k`.

    The computation of the splines is based on the scipy-package.

    Parameters
    ----------
    x : ndarray, shape (N,)
        `N` points to evaluate the B-spline at.
    knots : ndarray, shape >= (k+1,)
        Vector of knots derived from breaks (with appropriate endpoint
        multiplicity).
    order : int, positive
        Order `k` of the B-spline (4 = cubic).
    deriv: int, positive, optional
        Derivative of the B-spline partition (defaults to 0).

    Returns
    -------
    collmat : ndarray, shape (N, n-k)
        Collocation matrix, `n` is the size of `knots`.

    See Also
    --------
    augment_breaks

    """

    if deriv is None:
        deriv = 0

    if deriv >= order:
        return np.zeros((max(x.size, 1), knots.size - order))

    else:
        # create spline using scipy.interpolate
        coll = np.empty((max(x.size, 1), knots.size - order))

        for n in range(knots.size - order):
            c = np.zeros(knots.size - order)
            c[n] = 1.

            b = BSpline.construct_fast(knots, c, order-1, extrapolate=False)

            coll[:, n] = b(x, nu=deriv)

        return coll


def augment_breaks(breaks, order):
    """
    Augment a vector of break points and return the knot vector for a
    B-spline representation of order `k`.

    Parameters
    ----------
    breaks: ndarray, shape (n,)
        1-D array, containing `n` break points (without endpoint repeats).
    order: int, positive
        Order `k` of B-spline (4 = cubic).

    Returns
    -------
    knots: ndarray, shape (n+2k-2,)
        1-D array with `k`-times repeated copies of the `breaks`-vector
        endpoints ``breaks[0]`` and ``breaks[-1]``.

    """

    if isinstance(breaks, np.ndarray) and breaks.ndim > 1:
        raise ValueError("Breaks must be a 1-D array.")

    degree = order - 1
    return np.array([breaks[0]]*degree + list(breaks) + [breaks[-1]]*degree)


def pp_from_bspline(coeffs, knots, order):
    """
    Return a piecewise polynomial from a BSpline representation.

    Parameters
    ----------
    coeffs : ndarray, shape (M, D)
        Bspline coefficients for the `M` B-splines parameterizing
        `D` dimensions.
    knots : ndarray, shape (N,)
        B-spline knots. The knots must have the full endpoint multiplicity.
        Zero-pad spline coefficients if needed.
    order : int
        Order of the B-spline.

    Returns
    -------
    coeffs_pp : ndarray, shape (K, P, D)
        Coefficients of the piecewise polynomial where `K` is the order
        (``order``), `P` is the number of pieces and `D` is the dimension.
    breaks : ndarray, shape (P+1,)
        Break points of the piecewise polynomial.

    """

    degree = order - 1
    breaks = np.unique(knots)
    pieces = breaks.size - 1
    dim = coeffs.shape[-1]

    coeffs_pp = np.empty((order, pieces, dim))

    # have to do it manually for each dimension
    for d in range(dim):

        bs = BSpline(knots, coeffs[:, d], degree)
        pp = PPoly.from_spline(bs, extrapolate=False)

        # remove endpoint multiplicities
        coeffs_pp[:, :, d] = pp.c[:, degree:(degree+pieces)]

    return coeffs_pp, breaks


def synth_from_pp(breaks, order, coeffs, time, radius, theta, phi, *,
                  nmax=None, source=None, deriv=None, grid=None):
    """
    Compute radial, colatitude and azimuthal field components from the magnetic
    potential in terms of a spherical harmonic expansion in form of a
    piecewise polynomial.

    Parameters
    ----------
    breaks : ndarray, shape (m+1,)
        1-D array, containing `m+1` break points (without endpoint repeats) for
        `m` intervals.
    order : int, positive
        Order `k` of piecewise polynomials (4 = cubic).
    coeffs : ndarray, shape (k, m, nmax*(nmax+2))
        Coefficients of the piecewise polynomials, where `m` is the number of
        polynomial pieces. The trailing dimension is equal to the number of
        expansion coefficients for each interval.
    time : ndarray, shape (...)
        Array containing the time in days.
    radius : ndarray, shape (...) or float
        Array containing the radius in kilometers.
    theta : ndarray, shape (...) or float
        Array containing the colatitude in degrees
        :math:`[0^\\circ,180^\\circ]`.
    phi : ndarray, shape (...) or float
        Array containing the longitude in degrees.
    nmax : int, positive, optional
        Maximum degree harmonic expansion (default is given by ``coeffs``,
        but can also be smaller, if specified).
    source : {'internal', 'external'}, optional
        Magnetic field source (default is an internal source).
    deriv : int, positive, optional
        Derivative to be taken (default is 0).
    grid : bool, optional
        If ``True``, field components are computed on a regular grid. Arrays
        ``theta`` and ``phi`` must have one dimension less than the output grid
        since the grid will be created as their outer product.

    Returns
    -------
    B_radius, B_theta, B_phi : ndarray, shape (...)
        Radial, colatitude and azimuthal field components.

    See Also
    --------
    synth_values

    """

    # handle optional argument: nmax
    nmax_coeffs = int(np.sqrt(coeffs.shape[-1] + 1) - 1)  # degree for coeffs
    if nmax is None:
        nmax = nmax_coeffs
    elif nmax > nmax_coeffs:
        warnings.warn('Input nmax = {0} is incompatible with the number of '
                      'model coefficients. Using nmax = {1} instead.'.format(
                        nmax, nmax_coeffs))
        nmax = nmax_coeffs

    # handle optional argument: source
    source = 'internal' if source is None else source

    # compute SH coefficients from pp-form and take derivatives if needed
    deriv = 0 if deriv is None else deriv

    # set grid option to false
    grid = False if grid is None else grid

    PP = PPoly.construct_fast(coeffs[..., :nmax*(nmax+2)], breaks,
                              extrapolate=False)

    PP = PP.derivative(nu=deriv)
    gauss_coeffs = PP(time) * 365.25**deriv

    B_radius, B_theta, B_phi = synth_values(
        gauss_coeffs, radius, theta, phi, nmax=nmax, source=source, grid=grid)

    return B_radius, B_theta, B_phi


def synth_values(coeffs, radius, theta, phi, *, nmax=None, nmin=None,
                 mmax=None, source=None, grid=None):
    """
    Computes radial, colatitude and azimuthal field components from the
    magnetic potential field in terms of spherical harmonic coefficients.

    Parameters
    ----------

    coeffs : ndarray, shape (..., M)
        Coefficients of the spherical harmonic expansion. The last dimension is
        equal to the number of coefficients.
    radius : float or ndarray, shape (...)
        Array containing the radius in kilometers.
    theta : float or ndarray, shape (...)
        Array containing the colatitude in degrees
        :math:`[0^\\circ,180^\\circ]`.
    phi : float or ndarray, shape (...)
        Array containing the longitude in degrees.
    nmax : int, positive, optional
        Maximum degree up to which expansion is to be used (default is given by
        the last dimension of ``coeffs``, that is, ``M = nmax(nmax+2)``). This
        value can also be smaller if only coefficients at low degree should
        contribute.
    nmin : int, positive, optional
        Minimum degree of the expansion (defaults to 1). This value must
        correspond to the minimum degree of the coefficients and cannot be
        larger.
    mmax : int, positive, optional
        Maximum order of the spherical harmonic expansion (defaults to
        ``nmax``). This value can also be smaller. For example, if only the
        zonal part of the expansion should be used, set ``mmax = 0``.
    source : {'internal', 'external'}, optional
        Magnetic field source (default is an internal source).
    grid : bool, optional
        If ``True``, field components are computed on a regular grid. Arrays
        ``theta`` and ``phi`` must have one dimension less than the output grid
        since the grid will be created as their outer product (defaults to
        ``False``).

    Returns
    -------
    B_radius, B_theta, B_phi : ndarray, shape (...)
        Radial, colatitude and azimuthal field components.

    Warnings
    --------
    The function can also evaluate the field components at the geographic
    poles, i.e. where ``theta == 0.`` or ``theta == 180.``. However, users
    should be careful when doing this because the vector basis for spherical
    geocentric coordinates,
    :math:`{{\\mathbf{e}_r, \\mathbf{e}_\\theta, \\mathbf{e}_\\phi}}`,
    depends on longitude, which is not well defined at the poles. That is,
    at the poles, any value for the longitude maps to the same location in
    euclidean coordinates but gives a different vector basis in spherical
    geocentric coordinates. Nonetheless, by choosing a specific value for the
    longitude at the poles, users can define the vector basis, which then
    establishes the meaning of the spherical geocentric components returned by
    this function. The vector basis for the horizontal components is defined as

    .. math::

        \\mathbf{e}_\\theta &= \\cos\\theta\\cos\\phi\\mathbf{e}_x -
            \\cos\\theta\\sin\\phi\\mathbf{e}_y - \\sin\\theta\\mathbf{e}_z\\\\
        \\mathbf{e}_\\phi &= -\\sin\\phi\\mathbf{e}_x +
            \\cos\\phi\\mathbf{e}_y

    Hence, at the geographic north pole as given by ``theta = 0.`` and
    ``phi = 0.`` (chosen by the user), the returned component ``B_theta``
    will be along the direction :math:`\\mathbf{e}_\\theta = \\mathbf{e}_x` and
    ``B_phi`` along :math:`\\mathbf{e}_\\phi = \\mathbf{e}_y`. However,
    if ``phi = 180.`` is chosen, ``B_theta``
    will be along :math:`\\mathbf{e}_\\theta = -\\mathbf{e}_x` and ``B_phi``
    along :math:`\\mathbf{e}_\\phi = -\\mathbf{e}_y`.

    Notes
    -----
    The function can work with different grid shapes, but the inputs have to
    satisfy NumPy's `broadcasting rules \\
    <https://docs.scipy.org/doc/numpy-1.15.0/user/basics.broadcasting.html>`_
    (``grid=False``, default). This also applies to the dimension of the
    coefficients ``coeffs`` excluding the last dimension.

    The optional parameter ``grid`` is for convenience. If set to ``True``,
    a singleton dimension is appended (prepended) to ``theta`` (``phi``)
    for broadcasting to a regular grid. The other inputs ``radius`` and
    ``coeffs`` must then be broadcastable as before but now with the resulting
    regular grid.

    Examples
    --------
    The most straight forward computation uses a fully specified grid. For
    example, compute the magnetic field at :math:`N=50` grid points on the
    surface.

    .. code-block:: python

      import chaosmagpy.model_utils as cpm
      import numpy as np

      N = 50
      coeffs = np.ones((3,))  # degree 1 coefficients for all points
      radius = 6371.2 * np.ones((N,))  # radius of 50 points in km
      phi = np.linspace(-180., 180., num=N)  # azimuth of 50 points in deg.
      theta = np.linspace(0., 180., num=N)  # colatitude of 50 points in deg.

      B = cpm.synth_values(coeffs, radius, theta, phi)
      print([B[num].shape for num in range(3)])  # output shape (N,)

    Instead of `N` points, compute the field on a regular
    :math:`N\\times N`-grid in azimuth and colatitude (slow).

    .. code-block:: python

      radius_grid = 6371.2 * np.ones((N, N))
      phi_grid, theta_grid = np.meshgrid(phi, theta)  # grid of shape (N, N)

      B = cpm.synth_values(coeffs, radius_grid, theta_grid, phi_grid)
      print([B[num].shape for num in range(3)])  # output shape (N, N)

    But this is slow since some computations on the grid are repeatedly done.
    The preferred method is to use NumPy's broadcasting rules (fast).

    .. code-block:: python

      radius_grid = 6371.2  # float, () or (1,)-shaped array broadcasted to NxN
      theta_grid = theta[:, None]  # append singleton: (N, 1)
      phi_grid = phi[None, :]  # prepend singleton: (1, N)

      B = cpm.synth_values(coeffs, radius_grid, theta_grid, phi_grid)
      print([B[num].shape for num in range(3)])  # output shape (N, N)

    For convenience, you can do the same by using ``grid=True`` option, which
    appends a singleton dimension to ``theta`` and inserts a singleton
    dimension at the second to last position in the shape of ``phi``.

    .. code-block:: python

      B = cpm.synth_values(coeffs, radius_grid, theta, phi, grid=True)
      print([B[num].shape for num in range(3)])  # output shape (N, N)

    Remember that ``grid=False`` (default) will result in
    (N,)-shaped outputs as in the first example.

    """

    # ensure ndarray inputs
    coeffs = np.asarray(coeffs, dtype=float)
    radius = (np.asarray(radius, dtype=float)
              / config_utils.basicConfig['params.r_surf'])
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    theta_min = np.amin(theta)
    theta_max = np.amax(theta)

    if (theta_min <= 0.0) or (theta_max >= 180.0):
        if (theta_min == 0.0) or (theta_max == 180.0):
            warnings.warn('Input coordinates include the poles.')
            poles = True
        else:
            raise ValueError('Colatitude outside bounds [0, 180].')
    else:
        poles = False

    nmin = 1 if nmin is None else int(nmin)
    assert nmin > 0, 'The value of "nmin" must be greater than zero.'

    dim = coeffs.shape[-1]

    # handle nmax and mmax
    if (nmax is None) and (mmax is None):
        nmax = int(np.sqrt(dim + nmin**2) - 1)
        mmax = nmax

    elif mmax is None:  # but nmax given
        mmax = nmax

    elif nmax is None:  # but mmax given

        if mmax >= (nmin-1):
            nmax = int((dim - mmax*(mmax+2) + nmin**2 - 1) / (2*mmax+1) + mmax)

        else:
            nmax = int(dim / (2*mmax+1) + nmin - 1)

    if nmax < nmin:
        raise ValueError(f'Nothing to compute: nmax ({nmax}) < nmin ({nmin}).')

    # handle optional argument: source
    if source is None:
        source = 'internal'

    # handle grid option
    grid = False if grid is None else grid

    # manually broadcast input grid on surface
    if grid:
        theta = np.expand_dims(theta, axis=-1)  # append singleton dimension
        phi = np.expand_dims(phi, axis=-2)  # insert singleton before last

    # get shape of broadcasted result
    try:
        b = np.broadcast(
            radius, theta, phi, np.broadcast_to(0, coeffs.shape[:-1]))

    except ValueError:
        print('Cannot broadcast grid shapes (excl. last dimension of coeffs):')
        print(f'radius: {radius.shape}')
        print(f'theta:  {theta.shape}')
        print(f'phi:    {phi.shape}')
        print(f'coeffs: {coeffs.shape[:-1]}')
        raise

    grid_shape = b.shape

    # initialize radial dependence given the source
    if source == 'internal':
        r_n = radius**(-(nmin+2))
    elif source == 'external':
        r_n = radius**(nmin-1)
    else:
        raise ValueError("Source must be either 'internal' or 'external'.")

    # compute associated Legendre polynomials as (n, m, theta-points)-array
    Pnm = legendre_poly(nmax, theta)

    # save sinth for fast access
    sinth = Pnm[1, 1]

    # find the poles
    if poles:
        where_poles = (theta == 0.) | (theta == 180.)

    # convert phi to radians
    phi = np.radians(phi)

    # allocate arrays in memory
    B_radius = np.zeros(grid_shape)
    B_theta = np.zeros(grid_shape)
    B_phi = np.zeros(grid_shape)

    num = 0
    for n in range(nmin, nmax+1):

        if source == 'internal':
            B_radius += (n+1) * Pnm[n, 0] * r_n * coeffs[..., num]
        else:
            B_radius += -n * Pnm[n, 0] * r_n * coeffs[..., num]

        B_theta += -Pnm[0, n+1] * r_n * coeffs[..., num]

        num += 1
        for m in range(1, min(n, mmax)+1):

            cmp = np.cos(m*phi)
            smp = np.sin(m*phi)

            if source == 'internal':
                B_radius += ((n+1) * Pnm[n, m] * r_n
                             * (coeffs[..., num] * cmp
                                + coeffs[..., num+1] * smp))
            else:
                B_radius += -(n * Pnm[n, m] * r_n
                              * (coeffs[..., num] * cmp
                                 + coeffs[..., num+1] * smp))

            B_theta += (-Pnm[m, n+1] * r_n
                        * (coeffs[..., num] * cmp
                           + coeffs[..., num+1] * smp))

            # need special treatment at the poles because Pnm/sinth = 0/0 for
            # theta in {0., 180.},
            # use L'Hopital's rule to take the limit:
            # lim Pnm/sinth = k*dPnm | theta in {0., 180}
            # (evaluated at poles, where k=1 for theta=0 and k=-1 for
            # theta=180.); k = costh = P[1, 0] at poles for convenience
            if poles:
                div_Pnm = np.where(where_poles, Pnm[m, n+1], Pnm[n, m])
                div_Pnm[where_poles] *= Pnm[1, 0][where_poles]
                div_Pnm[~where_poles] /= sinth[~where_poles]
            else:
                div_Pnm = Pnm[n, m] / sinth

            B_phi += (m * div_Pnm * r_n
                      * (coeffs[..., num] * smp
                         - coeffs[..., num+1] * cmp))

            num += 2

        if source == 'internal':
            r_n = r_n / radius  # equivalent to r_n = radius**(-(n+2))
        else:
            r_n = r_n * radius  # equivalent to r_n = radius**(n-1)

    return B_radius, B_theta, B_phi


def design_gauss(radius, theta, phi, nmax, *, nmin=None, mmax=None,
                 source=None):
    """
    Computes matrices to connect the radial, colatitude and azimuthal field
    components to the magnetic potential field in terms of spherical harmonic
    coefficients (Schmidt quasi-normalized).

    Parameters
    ----------

    radius : ndarray, shape (...)
        Array containing the radius in kilometers.
    theta : ndarray, shape (...)
        Array containing the colatitude in degrees
        :math:`[0^\\circ,180^\\circ]`.
    phi : ndarray, shape (...)
        Array containing the longitude in degrees.
    nmax : int, positive
        Maximum degree of the sphercial harmonic expansion.
    nmin : int, positive, optional
        Minimum degree of the sphercial harmonic expansion (defaults to 1).
    mmax : int, positive, optional
        Maximum order of the spherical harmonic expansion (defaults to
        ``nmax``). For ``mmax = 0``, for example, only the zonal terms are
        returned.
    source : {'internal', 'external'}, optional
        Magnetic field source (default is an internal source).

    Returns
    -------
    A_radius, A_theta, A_phi : ndarray, shape (..., M)
        Matrices for radial, colatitude and azimuthal field components. The
        second dimension ``M`` varies depending on ``nmax``, ``nmin`` and
        ``mmax``.

    Warnings
    --------
    The function can also return the design matrix for the field components at
    the geographic poles, i.e. where ``theta == 0.`` or ``theta == 180.``.
    However, users should be careful when doing this because the vector basis
    for spherical geocentric coordinates,
    :math:`{{\\mathbf{e}_r, \\mathbf{e}_\\theta, \\mathbf{e}_\\phi}}`,
    depends on longitude, which is not well defined at the poles. That is,
    at the poles, any value for the longitude maps to the same location in
    euclidean coordinates but gives a different vector basis in spherical
    geocentric coordinates. Nonetheless, by choosing a specific value for the
    longitude at the poles, users can define the vector basis, which then
    establishes the meaning of the spherical geocentric components. The vector
    basis for the horizontal components is defined as

    .. math::

        \\mathbf{e}_\\theta &= \\cos\\theta\\cos\\phi\\mathbf{e}_x -
            \\cos\\theta\\sin\\phi\\mathbf{e}_y - \\sin\\theta\\mathbf{e}_z\\\\
        \\mathbf{e}_\\phi &= -\\sin\\phi\\mathbf{e}_x +
            \\cos\\phi\\mathbf{e}_y

    Hence, at the geographic north pole as given by ``theta = 0.`` and
    ``phi = 0.`` (chosen by the user), the returned design matrix ``A_theta``
    will be for components along the direction
    :math:`\\mathbf{e}_\\theta = \\mathbf{e}_x` and ``A_phi`` for components
    along :math:`\\mathbf{e}_\\phi = \\mathbf{e}_y`. However,
    if ``phi = 180.`` is chosen, ``A_theta`` will be for components along
    :math:`\\mathbf{e}_\\theta = -\\mathbf{e}_x` and ``A_phi``
    along :math:`\\mathbf{e}_\\phi = -\\mathbf{e}_y`.

    Examples
    --------
    Create the design matrices given 4 locations on the Earth's surface:

    >>> r = 6371.2
    >>> theta = np.array([90., 100., 110., 120.])
    >>> phi = 0.
    >>> A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=1)
    >>> A_radius
    array([[ 1.22464680e-16, 2.00000000e+00, 0.00000000e+00],
           [-3.47296355e-01, 1.96961551e+00, 0.00000000e+00],
           [-6.84040287e-01, 1.87938524e+00, 0.00000000e+00],
           [-1.00000000e+00, 1.73205081e+00, 0.00000000e+00]])

    Say, only the axial dipole coefficient is non-zero, what is the magnetic
    field in terms of spherical geocentric components?

    >>> m = np.array([-30000, 0., 0.])  # model coefficients in nT
    >>> Br = A_radius @ m; Br
    array([-3.67394040e-12, 1.04188907e+04, 2.05212086e+04, 3.00000000e+04])
    >>> Bt = A_theta @ m; Bt
    array([-30000. , -29544.23259037, -28190.77862358, -25980.76211353])
    >>> Bp = A_phi @ m; Bp
    array([0., 0., 0., 0.])

    A more complete example is given below:

    .. code-block:: python

        import numpy as np
        from chaosmagpy.model_utils import design_gauss, synth_values

        nmax = 5  # desired truncation degree, i.e. 35 model coefficients
        coeffs = np.arange(35)  # example model coefficients

        # example locations
        N = 10
        radius = 6371.2  # Earth's surface
        phi = np.linspace(-180., 180., num=N)  # azimuth in degrees
        theta = np.linspace(1., 179., num=N)  # colatitude in degrees

        # compute design matrices to compute predictions of the
        # field components from the model coefficients, each is of shape N x 35
        A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=nmax)

        # magnetic components computed from the model
        Br_pred = A_radius @ coeffs
        Bt_pred = A_theta @ coeffs
        Bp_pred = A_phi @ coeffs

        # compute the magnetic components directly from the coefficients
        Br, Bt, Bp = synth_values(coeffs, radius, theta, phi)
        np.linalg.norm(Br - Br_pred)  # approx 0.
        np.linalg.norm(Bt - Bt_pred)  # approx 0.
        np.linalg.norm(Bp - Bp_pred)  # approx 0.

    """

    # ensure ndarray inputs
    radius = (np.asarray(radius, dtype=float)
              / config_utils.basicConfig['params.r_surf'])
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # get shape of broadcasted result
    try:
        b = np.broadcast(radius, theta, phi)

    except ValueError:
        print('Cannot broadcast grid shapes:')
        print(f'radius: {radius.shape}')
        print(f'theta:  {theta.shape}')
        print(f'phi:    {phi.shape}')
        raise

    shape = b.shape

    theta_min = np.amin(theta)
    theta_max = np.amax(theta)

    # check if poles are included
    if (theta_min <= 0.0) or (theta_max >= 180.0):
        if (theta_min == 0.0) or (theta_max == 180.0):
            warnings.warn('Input coordinates include the poles.')
            poles = True
        else:
            raise ValueError('Colatitude outside bounds [0, 180].')
    else:
        poles = False

    # set internal source as default
    if source is None:
        source = 'internal'

    assert nmax > 0, '"nmax" must be greater than zero.'

    nmin = 1 if nmin is None else int(nmin)
    assert nmin <= nmax, '"nmin" must be smaller than "nmax".'
    assert nmin > 0, '"nmin" must be greater than zero.'

    mmax = nmax if mmax is None else int(mmax)
    assert mmax <= nmax, '"mmax" must be smaller than "nmax".'
    assert mmax >= 0, '"mmax" must be greater than or equal to zero.'

    # initialize radial dependence given the source
    if source == 'internal':
        r_n = radius**(-(nmin+2))
    elif source == 'external':
        r_n = radius**(nmin-1)
    else:
        raise ValueError("Source must be either 'internal' or 'external'.")

    # compute associated Legendre polynomials as (n, m, theta-points)-array
    Pnm = legendre_poly(nmax, theta)
    sinth = Pnm[1, 1]

    phi = np.radians(phi)

    # find the poles
    if poles:
        where_poles = (theta == 0.) | (theta == 180.)

    # compute the number of dimensions based on nmax, nmin, mmax
    if mmax >= (nmin-1):
        dim = int(mmax*(mmax+2) + (nmax-mmax)*(2*mmax+1) - nmin**2 + 1)
    else:
        dim = int((nmax-nmin+1)*(2*mmax+1))

    # allocate A_radius, A_theta, A_phi in memeory
    A_radius = np.zeros(shape + (dim,))
    A_theta = np.zeros(shape + (dim,))
    A_phi = np.zeros(shape + (dim,))

    num = 0
    for n in range(nmin, nmax+1):

        if source == 'internal':
            A_radius[..., num] = (n+1) * Pnm[n, 0] * r_n
        else:
            A_radius[..., num] = -n * Pnm[n, 0] * r_n

        A_theta[..., num] = -Pnm[0, n+1] * r_n

        num += 1
        for m in range(1, min(n, mmax)+1):

            cmp = np.cos(m*phi)
            smp = np.sin(m*phi)

            if source == 'internal':
                A_radius[..., num] = (n+1) * Pnm[n, m] * r_n * cmp
                A_radius[..., num+1] = (n+1) * Pnm[n, m] * r_n * smp
            else:
                A_radius[..., num] = -n * Pnm[n, m] * r_n * cmp
                A_radius[..., num+1] = -n * Pnm[n, m] * r_n * smp

            A_theta[..., num] = -Pnm[m, n+1] * r_n * cmp
            A_theta[..., num+1] = -Pnm[m, n+1] * r_n * smp

            # need special treatment at the poles because Pnm/sinth = 0/0 for
            # theta in {0., 180.},
            # use L'Hopital's rule to take the limit:
            # lim Pnm/sinth = k*dPnm | theta in {0., 180}
            # (evaluated at poles, where k=1 for theta=0 and k=-1 for
            # theta=180.); k = costh = P[1, 0] at poles for convenience
            if poles:
                div_Pnm = np.where(where_poles, Pnm[m, n+1], Pnm[n, m])
                div_Pnm[where_poles] *= Pnm[1, 0][where_poles]
                div_Pnm[~where_poles] /= sinth[~where_poles]
            else:
                div_Pnm = Pnm[n, m] / sinth

            A_phi[..., num] = m * div_Pnm * r_n * smp
            A_phi[..., num+1] = -m * div_Pnm * r_n * cmp

            num += 2

        # update radial dependence given the source
        if source == 'internal':
            r_n = r_n / radius
        else:
            r_n = r_n * radius

    return A_radius, A_theta, A_phi


def legendre_poly(nmax, theta):
    """
    Returns associated Legendre polynomials :math:`P_n^m(\\cos\\theta)`
    (Schmidt quasi-normalized) and the derivative
    :math:`dP_n^m(\\cos\\theta)/d\\theta` evaluated at :math:`\\theta`.

    Parameters
    ----------
    nmax : int, positive
        Maximum degree of the spherical expansion.
    theta : ndarray, shape (...)
        Colatitude in degrees :math:`[0^\\circ, 180^\\circ]`
        of arbitrary shape.

    Returns
    -------
    Pnm : ndarray, shape (n, m, ...)
          Evaluated values and derivatives, grid shape is appended as trailing
          dimensions. :math:`P_n^m(\\cos\\theta)` := ``Pnm[n, m, ...]`` and
          :math:`dP_n^m(\\cos\\theta)/d\\theta` := ``Pnm[m, n+1, ...]``

    References
    ----------
    Based on Equations 26-29 and Table 2 in:

    Langel, R. A., "Geomagnetism - The main field", Academic Press, 1987,
    chapter 4

    """

    costh = np.cos(np.radians(theta))
    sinth = np.sqrt(1-costh**2)

    Pnm = np.zeros((nmax+1, nmax+2) + costh.shape)
    Pnm[0, 0] = 1.  # is copied into trailing dimensions
    Pnm[1, 1] = sinth  # write theta into trailing dimenions via broadcasting

    rootn = np.sqrt(np.arange(2 * nmax**2 + 1))

    # Recursion relations after Langel "The Main Field" (1987),
    # eq. (27) and Table 2 (p. 256)
    for m in range(nmax):
        Pnm_tmp = rootn[m+m+1] * Pnm[m, m]
        Pnm[m+1, m] = costh * Pnm_tmp

        if m > 0:
            Pnm[m+1, m+1] = sinth*Pnm_tmp / rootn[m+m+2]

        for n in range(m+2, nmax+1):
            d = n * n - m * m
            e = n + n - 1
            Pnm[n, m] = ((e * costh * Pnm[n-1, m] - rootn[d-e] * Pnm[n-2, m])
                         / rootn[d])

    # dP(n,m) = Pnm(m,n+1) is the derivative of P(n,m) vrt. theta
    Pnm[0, 2] = -Pnm[1, 1]
    Pnm[1, 2] = Pnm[1, 0]
    for n in range(2, nmax+1):
        Pnm[0, n+1] = -np.sqrt((n*n + n) / 2) * Pnm[n, 1]
        Pnm[1, n+1] = ((np.sqrt(2 * (n*n + n)) * Pnm[n, 0]
                       - np.sqrt((n*n + n - 2)) * Pnm[n, 2]) / 2)

        for m in range(2, n):
            Pnm[m, n+1] = (0.5*(np.sqrt((n + m) * (n - m + 1)) * Pnm[n, m-1]
                           - np.sqrt((n + m + 1) * (n - m)) * Pnm[n, m+1]))

        Pnm[n, n+1] = np.sqrt(2 * n) * Pnm[n, n-1] / 2

    return Pnm


def power_spectrum(coeffs, radius=None, *, nmax=None, source=None, axis=None):
    """
    Compute the spatial power spectrum.

    Parameters
    ----------
    coeffs : ndarray, shape (..., N*(N+2))
        Spherical harmonic coefficients for degree `N`.
    radius : float, optional
        Radius in kilometers (defaults to mean Earth's surface radius).
        It has no effect for ``source='toroidal'``.
    nmax : int, optional
        Maximum spherical degree (defaults to `N`).
    source : {'internal', 'external', 'toroidal'}
        Source of the field model (defaults to internal).
    axis : int, optional
        Axis of ``coeffs`` along which to compute the spatial power spectrum
        (defaults to -1, last dimension of ``coeffs``).

    Returns
    -------
    W_n : ndarray, shape (..., ``nmax``)
        Power spectrum for degrees up to degree ``nmax``

    Notes
    -----
    The spatial power spectrum for a potential field is defined as

    .. math::

        W_n(r) &= \\langle|\\mathbf{B}|^2\\rangle
             = \\frac{1}{4\\pi}\\iint_\\Omega |\\mathbf{B}|^2 \\mathrm{d}
               \\Omega \\\\
             &= W_n^\\mathrm{i}(r) + W_n^\\mathrm{e}(r) + W_n^\\mathrm{T}

    where the internal :math:`W_n^\\mathrm{i}`, external
    :math:`W_n^\\mathrm{e}` and the non-potential toroidal
    :math:`W_n^\\mathrm{T}` spatial power spectra are

    .. math::

        W_n^\\mathrm{i}(r) &= (n+1)\\left(\\frac{a}{r}\\right)^{2n+4}
                              \\sum_{m=0}^n [(g_n^m)^2 + (h_n^m)^2] \\\\
        W_n^\\mathrm{e}(r) &= n\\left(\\frac{r}{a}\\right)^{2n-2}\\sum_{m=0}^n
                              [(q_n^m)^2 + (s_n^m)^2] \\\\
        W_n^\\mathrm{T}    &= \\frac{n(n+1)}{2n+1}\\sum_{m=0}^n
                              [(T_n^{m,c})^2 + (T_n^{m,s})^2]

    References
    ----------
    Sabaka, T. J.; Hulot, G. & Olsen, N.,
    "Mathematical properties relevant to geomagnetic field modeling",
    Handbook of geomathematics, Springer, 2010, 503-538

    """

    axis = -1 if axis is None else int(axis)
    if radius is None:
        r = 1
    else:
        r = config_utils.basicConfig['params.r_surf'] / radius

    N = int(np.sqrt(coeffs.shape[axis] + 1) - 1)  # maximum degree

    if nmax is None:
        nmax = N
    elif nmax > N:
        print(f'Incompatible maximum degree nmax = {nmax}, '
              f'setting nmax to {N}.')
        nmax = N

    source = 'internal' if source is None else source

    if source == 'internal':
        def factor(n, r):
            return (n+1) * r**(2*n+4)
    elif source == 'external':
        def factor(n, r):
            return n * r**(-(2*n-2))
    elif source == 'toroidal':
        def factor(n, _):
            return n*(n+1) / (2*n+1)
    else:
        raise ValueError(
            'Unknown source. Use `internal`, `external` or `toroidal`.')

    shape = list(coeffs.shape)
    shape[axis] = nmax  # replace dim of axis with nmax
    W_n = np.empty(shape)

    for n in range(1, nmax+1):
        min = n**2 - 1
        max = min + (2*n + 1)

        slc1 = coeffs.ndim*[slice(None), ]  # index for summation
        slc1[axis] = slice(min, max)

        slc2 = coeffs.ndim*[slice(None), ]  # index for insertion into output
        slc2[axis] = n-1

        W_n[tuple(slc2)] = factor(n, r)*np.sum(coeffs[tuple(slc1)]**2,
                                               axis=axis)

    return W_n


def degree_correlation(coeffs_1, coeffs_2):
    """
    Correlation per spherical harmonic degree between two models 1 and 2.

    Parameters
    ----------
    coeffs_1, coeffs_2 : ndarray, shape (N,)
        Two sets of coefficients of equal length `N`.

    Returns
    -------
    C_n : ndarray, shape (nmax,)
        Degree correlation of the two models. There are `N = nmax(nmax+2)`
        coefficients.

    """

    if coeffs_1.ndim != 1:
        raise ValueError(f'Only 1-D input allowed ({coeffs_1.ndim} != 1)')

    if coeffs_2.ndim != 1:
        raise ValueError(f'Only 1-D input allowed ({coeffs_2.ndim} != 1)')

    if coeffs_1.size != coeffs_2.size:
        raise ValueError(
            'Number of coefficients must be the same '
            '({0} != {1}).'.format(coeffs_1.size, coeffs_2.size))

    nmax = int(np.sqrt(coeffs_1.size + 1) - 1)

    C_n = np.zeros((nmax,))
    R_n = np.zeros((nmax,))  # elements are prop. to power spectrum of coeffs_1
    S_n = np.zeros((nmax,))  # elements are prop. to power spectrum of coeffs_2

    coeffs_12 = coeffs_1*coeffs_2

    for n in range(1, nmax+1):
        min = n**2 - 1
        max = min + (2*n + 1)
        R_n[n-1] = np.sum(coeffs_1[min:max]**2)
        S_n[n-1] = np.sum(coeffs_2[min:max]**2)
        C_n[n-1] = (np.sum(coeffs_12[min:max]) / np.sqrt(R_n[n-1]*S_n[n-1]))

    return C_n


def sensitivity(coeffs, coeffs_true):
    """
    Sensitivity (degree normalized error) of the input coefficients with
    respect to the target/true coefficients.

    Parameters
    ----------
    coeffs: ndarray, shape (N,)
        Input spherical harmonic coefficients.
    coeffs_true: ndarray, shape (N,)
        Target/true spherical harmonic coefficients.

    Returns
    -------
    S: ndarray, shape (N,)
        Sensitivity of ``coeffs`` with respect to ``coeffs_true``.

    Examples
    --------
    >>> import chaosmagpy as cp
    >>> import numpy as np
    >>> coeffs = np.array([1., 2., 5.])  # estimated degree 1 coefficients
    >>> coeffs_true = np.array([1., 2., 3.])  # target degree 1 coefficients
    >>> cp.model_utils.sensitivity(coeffs, coeffs_true)
    array([0., 0., 0.63245553])

    """
    if coeffs.ndim != 1:
        raise ValueError(f'Only 1-D input allowed ({coeffs.ndim} != 1)')

    if coeffs_true.ndim != 1:
        raise ValueError(f'Only 1-D input allowed ({coeffs_true.ndim} != 1)')

    if coeffs.size != coeffs_true.size:
        raise ValueError(
            'Number of coefficients must be the same '
            '({0} != {1}).'.format(coeffs.size, coeffs_true.size))

    nmax = int(np.sqrt(coeffs.size + 1) - 1)

    S = coeffs - coeffs_true

    # apply degree-dependent scaling
    for n in range(1, nmax+1):
        min = n**2 - 1
        max = min + (2*n + 1)

        S[min:max] *= 1. / np.sqrt(np.sum(coeffs_true[min:max]**2) / (2*n + 1))

    return S



# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

"""
`chaosmagpy.data_utils` provides functions for loading and writing data and
geomagnetic field models. It also offers functions to do common time
conversions.

.. autosummary::
    :toctree: functions

    load_matfile
    load_RC_datfile
    save_RC_h5file
    load_shcfile
    save_shcfile
    mjd2000
    timestamp
    is_leap_year
    dyear_to_mjd
    mjd_to_dyear
    memory_usage
    gauss_units

"""


import pandas as pd
import numpy as np
import hdf5storage as hdf
import warnings
import h5py
import os
import datetime as dt
import textwrap


def load_matfile(filepath, variable_names=None, **kwargs):
    """
    Load MAT-file and return dictionary.

    Function loads MAT-file by traversing the structure converting data into
    low-level numpy arrays of different types. There is no guarantee that any
    kind of data is read in correctly. The data dtype can also vary depending
    on the MAT-file (v7.3 returns floats instead of integers). But it should
    work identically for v7.3 and prior MAT-files. Arrays are squeezed if
    possible. Relies on the :mod:`hdf5storage` package.

    Parameters
    ----------
    filepath : str
        Filepath and name of MAT-file.
    variable_names : list of strings
        Top-level variables to be loaded.
    **kwargs : keywords
        Additional keyword arguments are passed to :func:`hdf5storage.loadmat`.

    Returns
    -------
    data : dict
        Dictionary containing the data as dictionaries or numpy arrays.

    """

    # define a recursively called function to traverse structure
    def traverse_struct(struct):

        # for dictionaries, iterate through keys
        if isinstance(struct, dict):
            out = dict()
            for key, value in struct.items():
                out[key] = traverse_struct(value)
            return out

        # for ndarray, iterate through dtype names
        elif isinstance(struct, np.ndarray):

            # collect dtype names if available
            names = struct.dtype.names

            # if no fields in array
            if names is None:
                if struct.dtype == np.dtype('O') and struct.shape == (1, 1):
                    return traverse_struct(struct[0, 0])
                else:
                    return struct.squeeze()

            else:  # if there are fields, iterate through fields
                out = dict()
                for name in names:
                    out[name] = traverse_struct(struct[name])
                return out

        else:
            return struct

    output = hdf.loadmat(filepath, variable_names=variable_names, **kwargs)

    # loadmat returns dictionary, go through keys and call traverse_struct
    for key, value in output.items():
        if key.startswith('__') and key.endswith('__'):
            pass
        else:
            output[key] = traverse_struct(value)

    return output


def load_RC_datfile(filepath=None, parse_dates=None):
    """
    Load RC-index data file into pandas data frame.

    Parameters
    ----------
    filepath : str, optional
        Filepath to RC index ``*.dat``. If ``None``, the RC
        index will be fetched from `spacecenter.dk <http://www.spacecenter.dk/\
        files/magnetic-models/RC/current/>`_.
    parse_dates : bool, optional
        Replace index with datetime object for time-series manipulations.
        Default is ``False``.

    Returns
    -------
    df : dataframe
        Pandas dataframe with names {'time', 'RC', 'RC_e', 'RC_i', 'flag'},
        where ``'time'`` is given in modified Julian dates.

    """

    if filepath is None:
        from lxml import html
        import requests

        link = "http://www.spacecenter.dk/files/magnetic-models/RC/current/"

        page = requests.get(link)
        print(f'Accessing {page.url}.')

        tree = html.fromstring(page.content)
        file = tree.xpath('//tr[5]//td[2]//a/@href')[0]  # get name from list
        date = tree.xpath('//tr[5]//td[3]/text()')[0]

        print(f'Downloading RC-index file "{file}" '
              f'(last modified on {date.strip()}).')

        filepath = link + file

    column_names = ['time', 'RC', 'RC_e', 'RC_i', 'flag']
    column_types = {'time': 'float64', 'RC': 'float64', 'RC_e': 'float64',
                    'RC_i': 'float64', 'flag': 'category'}

    df = pd.read_csv(filepath,  delim_whitespace=True, comment='#',
                     dtype=column_types, names=column_names)

    parse_dates = False if parse_dates is None else parse_dates

    # set datetime as index
    if parse_dates:
        df.index = pd.to_datetime(
            df['time'].values, unit='D', origin=pd.Timestamp('2000-1-1'))
        df.drop(['time'], axis=1, inplace=True)  # delete redundant time column

    return df


def save_RC_h5file(filepath, read_from=None):
    """
    Return HDF5-file of the RC index.

    Parameters
    ----------
    filepath : str
        Filepath and name of ``*.h5`` output file.
    read_from : str, optional
        Filepath of RC index ``*.dat``. If ``None``, the RC
        index will be fetched from :rc_url:`spacecenter.dk <>`.

    Notes
    -----
    Saves an HDF5-file of the RC index with keywords
    ['time', 'RC', 'RC_e', 'RC_i', 'flag']. Time is given in modified Julian
    dates 2000.

    Examples
    --------
    Save RC-index TXT-file (``RC_1997-2020_Aug_v4.dat``) as file in HDF5 format
    (``RC_index.h5``).

    >>> save_RC_h5file('RC_index.h5', read_from='RC_1997-2020_Aug_v4.dat')
    Successfully saved to RC_index.h5.

    """

    try:
        df_rc = load_RC_datfile(read_from, parse_dates=False)

        with h5py.File(filepath, 'w') as f:

            for column in df_rc.columns:
                variable = df_rc[column].values
                if column == 'flag':
                    dset = f.create_dataset(column, variable.shape, dtype="S1")
                    dset[:] = variable.astype('bytes')

                else:
                    f.create_dataset(column, data=variable)  # just save floats

            print(f'Successfully saved to {f.filename}.')

    except Exception as err:
        warnings.warn(f"Can't save new RC index. Raised exception: '{err}'.")


def load_shcfile(filepath, leap_year=None, comment=None):
    """
    Load SHC-file and return coefficient arrays.

    Parameters
    ----------
    filepath : str
        File path to spherical harmonic coefficient SHC-file.
    leap_year : {True, False}, optional
        Take leap year in time conversion into account (default). Otherwise,
        use conversion factor of 365.25 days per year.
    comment : str, optional
        Character at the start of a line to indicate a comment
        (defaults to ``#``). This can also be a tuple of characters.

    Returns
    -------
    time : ndarray, shape (N,)
        Array containing `N` times for each model snapshot in modified
        Julian dates with origin January 1, 2000 0:00 UTC.
    coeffs : ndarray, shape (nmax(nmax+2), N)
        Coefficients of model snapshots. Each column is a snapshot up to
        spherical degree and order `nmax`.
    parameters : dict, {'SHC', 'nmin', 'nmax', 'N', 'order', 'step'}
        Dictionary containing parameters of the model snapshots and the
        following keys: ``'SHC'`` SHC-file name, ``'nmin'`` minimum degree,
        ``'nmax'`` maximum degree, ``'N'`` number of snapshot models,
        ``'order'`` piecewise polynomial order and ``'step'`` number of
        snapshots until next break point. Extract break points of the
        piecewise polynomial with ``breaks = time[::step]``.

    """

    leap_year = True if leap_year is None else leap_year
    comment = '#' if comment is None else comment

    first_line = True

    with open(filepath, 'r') as f:

        data = np.array([])
        for line in f.readlines():

            if line.strip().startswith(comment):
                continue

            newline = np.fromstring(line, sep=' ')

            if first_line:  # first non-comment line contains shc params
                name = os.path.split(filepath)[1]  # file name string
                values = [name] + newline.astype(int).tolist()

                first_line = False

            else:
                data = np.append(data, newline)

        # unpack parameter line
        keys = ['SHC', 'nmin', 'nmax', 'N', 'order', 'step']
        parameters = dict(zip(keys, values))

        time = data[:parameters['N']]
        coeffs = data[parameters['N']:].reshape((-1, parameters['N']+2))
        coeffs = coeffs[:, 2:].copy()  # discard columns with n and m

        mjd = dyear_to_mjd(time, leap_year=leap_year)

    return mjd, coeffs, parameters


def save_shcfile(time, coeffs, order=None, filepath=None, nmin=None, nmax=None,
                 leap_year=None, header=None):
    """
    Save Gauss coefficients as SHC-file.

    Parameters
    ----------
    time : float, list, ndarray, shape (n,)
        Time of model coeffcients in modified Julian date.
    coeffs : ndarray, shape (N,) or (n, N)
        Gauss coefficients as vector or array. The first dimension of the array
        must be equal to the length `n` of the given ``time``.
    order : int, optional (defaults to 1)
        Order of the piecewise polynomial with which the coefficients are
        parameterized in time (breaks are given by ``time[::order]``).
    filepath : str, optional
        Filepath and name of the output file. Defaults to the current working
        directory and filename `model.shc`.
    nmin : int, optional
        Minimum spherical harmonic degree (defaults to 1). This will remove
        first values from coeffs if greater than 1.
    nmax : int, optional
        Maximum spherical harmonic degree (defaults to degree compatible with
        number of coeffcients, otherwise coeffcients are truncated).
    leap_year : {True, False}, optional
        Take leap years for decimal year conversion into account
        (defaults to ``True``).
    header : str, optional
        Optional header at beginning of file. Defaults to an empty string.

    """

    time = np.asarray(time, dtype=float)

    order = 1 if order is None else int(order)

    nmin = 1 if nmin is None else int(nmin)

    if nmax is None:
        nmax = int(np.sqrt(coeffs.shape[-1] + 1) - 1)
    else:
        nmax = int(nmax)

    if nmin > nmax:
        raise ValueError('``nmin`` must be smaller than or equal to ``nmax``.')

    filepath = 'model.shc' if filepath is None else filepath

    header = '' if header is None else header

    leap_year = True if leap_year is None else bool(leap_year)

    if coeffs.ndim == 1:
        coeffs = coeffs.reshape((1, -1))

    coeffs = coeffs[:, (nmin**2-1):((nmax+1)**2-1)]

    # compute all possible degree and orders
    deg = np.array([], dtype=int)
    ord = np.array([], dtype=int)
    for n in range(nmin, nmax+1):
        deg = np.append(deg, np.repeat(n, 2*n+1))
        ord = np.append(ord, [0])
        for m in range(1, n+1):
            ord = np.append(ord, [m, -m])

    comment = header + textwrap.dedent(f"""\
        # Created on {dt.datetime.now(dt.timezone.utc)} UTC.
        # Leap years are accounted for in decimal years format ({leap_year}).
        {nmin} {nmax} {time.size} {order} {order-1}
        """)

    with open(filepath, 'w') as f:
        # write comment line
        f.write(comment)

        f.write('  ')  # to represent two missing values
        for t in time:
            f.write(' {:16.8f}'.format(mjd_to_dyear(t, leap_year=leap_year)))
        f.write('\n')

        # write coefficient table to 8 significants
        for row, (n, m) in enumerate(zip(deg, ord)):

            f.write('{:} {:}'.format(n, m))

            for value in coeffs[:, row]:
                f.write(' {:16.8f}'.format(value))

            f.write('\n')

    print('Created SHC-file {}.'.format(
        os.path.join(os.getcwd(), filepath)))


def mjd2000(year, month=1, day=1, hour=0, minute=0, second=0, microsecond=0):
    """
    Computes the modified Julian date as floating point number (epoch 2000).

    It assigns 0 to 0h00 January 1, 2000. Leap seconds are not accounted for.

    Parameters
    ----------
    time : :class:`datetime.datetime`, ndarray, shape (...)
        Datetime class instance, `OR ...`
    year : int, ndarray, shape (...)
    month : int, ndarray, shape (...), optional
        Month of the year `[1, 12]` (defaults to 1).
    day : int, ndarray, shape (...), optional
        Day of the corresponding month (defaults to 1).
    hour : int , ndarray, shape (...), optional
        Hour of the day (default is 0).
    minute : int, ndarray, shape (...), optional
        Minutes (default is 0).
    second : int, ndarray, shape (...), optional
        Seconds (default is 0).
    microsecond : int, ndarray, shape (...), optional
        Microseconds (default is 0).

    Returns
    -------
    time : ndarray, shape (...)
        Modified Julian date (units of days).

    Examples
    --------
    >>> a = np.array([datetime.datetime(2000, 1, 1), \
datetime.datetime(2002, 3, 4)])
    >>> mjd2000(a)
        array([  0., 793.])

    >>> mjd2000(2003, 5, 3, 13, 52, 15)  # May 3, 2003, 13:52:15 (hh:mm:ss)
        1218.5779513888888

    >>> mjd2000(np.arange(2000, 2005))  # January 1 in each year
        array([   0.,  366.,  731., 1096., 1461.])

    >>> mjd2000(np.arange(2000, 2005), 2, 1)  # February 1 in each year
        array([  31.,  397.,  762., 1127., 1492.])

    >>> mjd2000(np.arange(2000, 2005), 2, np.arange(1, 6))
        array([  31.,  398.,  764., 1130., 1496.])

    """

    year = np.asarray(year)

    if (np.issubdtype(year.dtype, np.dtype(dt.datetime).type) or
            np.issubdtype(year.dtype, np.datetime64)):
        datetime = year.astype('datetime64[us]')

    else:
        # build iso datetime string with unicode
        year = np.asarray(year, dtype=np.unicode_)
        month = np.char.zfill(np.asarray(month, dtype=np.unicode_), 2)
        day = np.char.zfill(np.asarray(day, dtype=np.unicode_), 2)

        year_month = np.char.add(np.char.add(year, '-'), month)
        datetime = np.char.add(np.char.add(year_month, '-'), day)

        datetime = datetime.astype('datetime64[us]')

        # not use inplace add here because it doesn't broadcast arrays
        datetime = datetime + np.asarray(hour, dtype='timedelta64[h]')
        datetime = datetime + np.asarray(minute, dtype='timedelta64[m]')
        datetime = datetime + np.asarray(second, dtype='timedelta64[s]')
        datetime = datetime + np.asarray(microsecond, dtype='timedelta64[us]')

    microseconds = datetime - np.datetime64('2000-01-01', 'us')

    return microseconds / np.timedelta64(1, 'D')  # fraction of days


def timestamp(time):
    """
    Convert modified Julian date to NumPy's datetime format.

    Parameters
    ----------
    time : ndarray, shape (...)
        Modified Julian date (units of days).

    Returns
    -------
    time : ndarray, shape (...)
        Array of ``numpy.datetime64[us]``.

    Examples
    --------
    >>> timestamp(0.53245)
        numpy.datetime64('2000-01-01T12:46:43.680000')

    >>> timestamp(np.linspace(0., 1.5, 2))
        array(['2000-01-01T00:00:00.000000', '2000-01-02T12:00:00.000000'], \
dtype='datetime64[us]')

    """

    # convert mjd2000 to timedelta64[us]
    us = np.asarray(time) * 86400e6 * np.timedelta64(1, 'us')

    # add datetime offset
    return us + np.datetime64('2000-01-01', 'us')


def is_leap_year(year):
    """
    Determine if input year is a leap year.

    Parameters
    ----------
    year : int, ndarray, shape (...)
        Years to test for leap year.

    Returns
    -------
    leap_year : ndarray of bools, shape (...)
        ``True`` for leap year in array.

    Examples
    --------
    >>> is_leap_year([2000, 2001, 2004])
        array([ True, False,  True])

    Raises
    ------
    TypeError if ``year`` is not of type integer.

    """

    year = np.asarray(year)

    if not np.issubdtype(year.dtype, int):
        raise TypeError('Expected integer values as the input year. Use '
                        'numpy.floor to extract the integer year '
                        'from decimal years.')

    return np.logical_and(np.remainder(year, 4) == 0,
                          np.logical_or(np.remainder(year, 100) != 0,
                                        np.remainder(year, 400) == 0))


def dyear_to_mjd(time, leap_year=None):
    """
    Convert time from decimal years to modified Julian date 2000.

    Leap years are accounted for by default.

    Parameters
    ----------
    time : float, ndarray, shape (...)
        Time in decimal years.
    leap_year : {True, False}, optional
        Take leap years into account by using a conversion factor of 365 or 366
        days in a year (leap year, used by default). If ``False`` a conversion
        factor of 365.25 days in a year is used.

    Returns
    -------
    time : ndarray, shape (...)
        Time in modified Julian date 2000.

    Examples
    --------
    >>> dyear_to_mjd([2000.5, 2001.5])  # account for leap years
        array([183. , 548.5])

    >>> dyear_to_mjd([2000.5, 2001.5], leap_year=False)
        array([182.625, 547.875])

    """

    leap_year = True if leap_year is None else leap_year

    if leap_year:
        year = np.asarray(np.floor(time), dtype=int)  # note: -0.1 is year -1
        frac_of_year = np.remainder(time, 1.)

        isleap = is_leap_year(year)  # do provide integer years
        days_per_year = np.where(isleap, 366., 365.)

        days = frac_of_year * days_per_year

        mjd = mjd2000(year, 1, 1) + days

    elif not leap_year:
        days_per_year = 365.25

        mjd = (np.asarray(time) - 2000.0) * days_per_year

    else:
        raise ValueError('Unknown leap year option: use either True or False')

    return mjd


def mjd_to_dyear(time, leap_year=None):
    """
    Convert time from modified Julian date 2000 to decimal years.

    Leap years are accounted for by default.

    Parameters
    ----------
    time : float, ndarray, shape (...)
        Time in modified Julian date 2000.
    leap_year : {True, False}, optional
        Take leap years into account by using a conversion factor of 365 or 366
        days in a year (leap year, used by default). If ``False`` a conversion
        factor of 365.25 days in a year is used.

    Returns
    -------
    time : ndarray, shape (...)
        Time in decimal years.

    Examples
    --------
    >>> mjd_to_dyear([183. , 548.5])  # account for leap years
        array([2000.5, 2001.5])

    >>> mjd_to_dyear([0. , -1., 365.])
        array([2000., 1999.99726027, 2000.99726776])

    >>> mjd_to_dyear([182.625, 547.875], leap_year=False)
        array([2000.5, 2001.5])

    """

    leap_year = True if leap_year is None else leap_year

    if leap_year:
        date = (np.asarray(np.floor(time), dtype=int)*np.timedelta64(1, 'D')
                + np.datetime64('2000-01-01'))  # only precise to date

        year = date.astype('datetime64[Y]').astype(int) + 1970
        days = np.asarray(time) - mjd2000(year, 1, 1)  # days of that year

        isleap = is_leap_year(year)
        days_per_year = np.where(isleap, 366., 365.)

        dyear = year + days / days_per_year

    elif not leap_year:
        dyear = np.asarray(time) / 365.25 + 2000.

    else:
        raise ValueError('Unknown leap year option: use either True or False')

    return dyear


def memory_usage(pandas_obj):
    """
    Compute memory usage of pandas object.

    For full report, use: ``df.info(memory_usage='deep')``.

    """

    if isinstance(pandas_obj, pd.DataFrame):
        usage_b = pandas_obj.memory_usage(deep=True).sum()
    else:  # we assume if not a df it's a series
        usage_b = pandas_obj.memory_usage(deep=True)
    usage_mb = usage_b / 1024 ** 2  # convert bytes to megabytes
    return "{:03.2f} MB".format(usage_mb)


def gauss_units(deriv=None):
    """
    Return string of the magnetic field units given the derivative with time.

    String is meant to be used in plot labels.

    Parameters
    ----------
    deriv : int, optional
        Derivative (defaults to 0).

    Returns
    -------
    units : str
        Tex-style unit string.

    Examples
    --------
    >>> gauss_units()
        'nT'

    >>> gauss_units(1)
        '$\\\\mathrm{nT}/\\\\mathrm{yr}$'

    >>> gauss_units(2)
        '$\\\\mathrm{nT}/\\\\mathrm{yr}^{2}$'

    """

    deriv = 0 if deriv is None else deriv

    if deriv == 0:
        units = 'nT'
    elif deriv == 1:
        units = '$\\mathrm{{nT}}/\\mathrm{{yr}}$'
    else:
        units = '$\\mathrm{{nT}}/\\mathrm{{yr}}^{{{:}}}$'.format(deriv)

    return units



# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

"""
`chaosmagpy.config_utils` contains functions and classes to manipulate the
configuration dictionary, which contains parameters and options in ChaosMagPy.
The following list gives an overview of the possible keywords. The keywords can
be accessed after importing chaosmagpy through:

>>> import chaosmagpy as cp
>>> cp.basicConfig['params.r_surf']  # get Earth's mean radius in km
    6371.2

**Parameters**

 ======================  =============  =======================================
 Value                   Type           Description
 ======================  =============  =======================================
 'params.r_surf'         `float`        Reference radius in kilometers
                                        (defaults to Earth's surface radius of
                                        6371.2 km).
 'params.r_cmb'          `float`        Core-mantle boundary radius in
                                        kilometers (defaults to 3485.0 km).
 'params.dipole'         `list`,        Coefficients of the dipole (used for
                         `ndarray`,     GSM/SM coordinate transformations) in
                         `shape (3,)`   units of nanotesla.
 'params.ellipsoid'      `list`,        Equatorial (index 0) and polar radius
                         `ndarray`      (index 1) of the spheroid describing
                         `shape (2,)`   Earth (WGS84) in units of kilometers.
 'params.CHAOS_version'  `str`          Version of the CHAOS model that was
                                        constructed with the built-in RC-index.
 'params.cdf_to_mjd'     `int`          Number of days on Jan 01, 2000 since
                                        Jan 01, 0000 (CDF start epoch)
 ======================  =============  =======================================

**Files**

 ==========================  ============  ====================================
 Value                       Type          Description
 ==========================  ============  ====================================
 'file.RC_index'             `HDF5-file`,  RC-index file (used for external
                             `TXT-file`    field computation). See also
                                           :func:`data_utils.save_RC_h5file`.
 'file.GSM_spectrum'         `NPZ-file`    GSM transformation coefficients. See
                                           also :func:`coordinate_utils.\\
                                           rotate_gauss_fft`.
 'file.SM_spectrum'          `NPZ-file`    SM transformation coefficients. See
                                           also :func:`coordinate_utils.\\
                                           rotate_gauss_fft`.
 'file.Earth_conductivity'   `TXT-file`    Conductivity model of a layered
                                           Earth (deprecated).
 'file.shp_coastline'        `SHP-file`,   Coastline shapefile (default file
                             `ZIP-file`    provided by Natural Earth).
 ==========================  ============  ====================================

**Plots**

 ==========================  ===========  =====================================
 Value                       Type         Description
 ==========================  ===========  =====================================
 'plots.figure_width'        `float`      Plot width in inches (defaults to 6.3
                                          inches or 16 cm)
 ==========================  ===========  =====================================

.. autosummary::
    :toctree: classes
    :template: myclass.rst

    BasicConfig

"""

import os
import re
import json
import numpy as np
import warnings
from contextlib import contextmanager

ROOT = os.path.abspath(os.path.dirname(__file__))
LIB = os.path.join(ROOT, 'lib')


# copied/inspired by matplotlib.rcsetup
def check_path_exists(s):
    """Check that path to file exists."""
    if s is None or s == 'None':
        return None
    if os.path.exists(s):
        return s
    else:
        raise FileNotFoundError(f'{s} does not exist.')


def check_float(s):
    """Convert to float."""
    try:
        return float(s)
    except ValueError:
        raise ValueError(f'Could not convert {s} to float.')


def check_int(s):
    """Convert to integer."""
    try:
        return int(s)
    except ValueError:
        raise ValueError(f'Could not convert {s} to integer.')


def check_string(s):
    """Convert to string."""
    try:
        return str(s)
    except ValueError:
        raise ValueError(f'Could not convert {s} to string.')


def check_vector(s, len=None):
    """Check that input is vector of required length."""
    try:
        s = np.array(s)
        assert s.ndim == 1
        if len is not None:
            if s.size != len:
                raise ValueError(f'Wrong length: {s.size} != {len}.')
        return s
    except Exception as err:
        raise ValueError(f'Not a valid vector. {err}')


def check_version_string(s):
    """Check correct format of version string."""

    s = check_string(s)

    match = re.search(r'\d+\.\d+', s)
    if match:
        return s
    else:
        raise ValueError(f'Not supported version format "{s}".'
                         'Must be of the form "x.x" with x an integer.')


DEFAULTS = {
    'params.r_surf': [6371.2, check_float],
    'params.r_cmb': [3485.0, check_float],
    'params.dipole': [np.array([-29442.0, -1501.0, 4797.1]),
                      lambda x: check_vector(x, len=3)],
    'params.ellipsoid': [np.array([6378.137, 6356.752]),
                         lambda x: check_vector(x, len=2)],
    'params.CHAOS_version': ['7.17', check_version_string],
    'params.cdf_to_mjd': [730485, check_int],

    # location of coefficient files
    'file.RC_index': [os.path.join(LIB, 'RC_index.h5'),
                      check_path_exists],
    'file.GSM_spectrum': [os.path.join(LIB, 'frequency_spectrum_gsm.npz'),
                          check_path_exists],
    'file.SM_spectrum': [os.path.join(LIB, 'frequency_spectrum_sm.npz'),
                         check_path_exists],
    'file.Earth_conductivity': [os.path.join(LIB, 'Earth_conductivity.dat'),
                                check_path_exists],
    'file.shp_coastline': [os.path.join(LIB, 'ne_110m_coastline.zip'),
                           check_path_exists],

    # plot related configuration
    'plots.figure_width': [6.3, check_float],
}


class BasicConfig(dict):
    """Class for creating CHAOS configuration dictionary."""

    defaults = DEFAULTS

    def __init__(self, *args, **kwargs):
        super().update(*args, **kwargs)

    def __setitem__(self, key, value):
        """Set and check value before updating dictionary."""

        try:
            try:
                cval = self.defaults[key][1](value)
            except ValueError as err:
                raise ValueError(f'Key "{key}": {err}')
            super().__setitem__(key, cval)
        except KeyError:
            raise KeyError(f'"{key}" is not a valid parameter.')

    def __str__(self):
        return '\n'.join(map('{0[0]}: {0[1]}'.format, sorted(self.items())))

    def reset(self, key):
        """
        Load default values.

        Parameters
        ----------
        key : str
            Single keyword that is reset to the default.

        """
        self.__setitem__(key, self.defaults[key][0])

    def fullreset(self):
        """
        Load all default values.

        """
        super().update({key: val for key, (val, _) in self.defaults.items()})

    def load(self, filepath):
        """
        Load configuration dictionary from file.

        Parameters
        ----------
        filepath : str
            Filepath and name to json-formatted configuration txt-file.

        """

        with open(filepath, 'r') as f:
            kwargs = json.load(f)

        if len(kwargs) == 0:
            warnings.warn(
                'Configuration dictionary loaded from file is empty.')

        for key, value in kwargs.items():
            # check format and set key value pairs
            self.__setitem__(key, value)

    def save(self, filepath):
        """
        Save configuration dictionary to a file.

        Parameters
        ----------
        filepath : str
            Filepath and name of the textfile that will be saved with the
            configuration values.

        """

        def default(obj):
            if isinstance(obj, np.ndarray):
                return obj.tolist()

        with open(filepath, 'w') as f:
            json.dump(self, f, default=default, indent=4, sort_keys=True)

        print(f'Saved configuration textfile to {filepath}.')

    @contextmanager
    def context(self, key, value):
        """
        Use context manager to temporarily change setting.

        Parameters
        ----------
        key : str
            BasicConfig configuration key.
        value
            Value compatible with ``key``.

        Examples
        --------
        Temporarily change the radius of Earth's surface for a computation
        and then change it back to the original value.

        .. code-block:: python

          from chaosmagpy import basicConfig

          print('Before: ', basicConfig['params.r_surf'])

          # change Earth's radius to 10 km
          with basicConfig.context('params.r_surf', 10):
              # do something at r_surf = 10 km ...
              print('Inside: ', basicConfig['params.r_surf'])

          print('After: ', basicConfig['params.r_surf'])

        """
        old_value = self.__getitem__(key)
        self.__setitem__(key, value)
        yield
        self.__setitem__(key, old_value)


# load defaults
basicConfig = BasicConfig({key: val for key, (val, _) in DEFAULTS.items()})


if __name__ == '__main__':
    # ensure default passes tests
    for key, (value, test) in DEFAULTS.items():
        if not np.all(test(value) == value):
            print(f"{key}: {test(value)} != {value}")



# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

import numpy as np
import os
import textwrap
import chaosmagpy as cp
from chaosmagpy import coordinate_utils as cpc
from unittest import TestCase, main
from math import pi
from timeit import default_timer as timer

try:
    from tests.helpers import load_matfile
except ImportError:
    from helpers import load_matfile

ROOT = os.path.abspath(os.path.dirname(__file__))
MATFILE_PATH = os.path.join(ROOT, 'data/CHAOS_test.mat')

# check if mat-file exists in tests directory
if os.path.isfile(MATFILE_PATH) is False:
    MATFILE_PATH = str(input('Matfile path for coordinate_utils test?: '))


class CoordinateUtils(TestCase):
    def setUp(self):

        print(textwrap.dedent(f"""\

            {"":-^70}
            Running {self._testMethodName}:
            """))

    def test_dipole_to_unit(self):

        np.testing.assert_allclose(
            cpc.igrf_dipole('2010'),
            cpc._dipole_to_unit(11.32, 289.59))

        np.testing.assert_allclose(
            cpc.igrf_dipole('2015'),
            cpc._dipole_to_unit(np.array([-29442.0, -1501.0, 4797.1])))

        np.testing.assert_allclose(
            cpc.igrf_dipole('2015'),
            cpc._dipole_to_unit(-29442.0, -1501.0, 4797.1))

        # test higher-dimensional arrays

        desired = np.tile(cpc.igrf_dipole('2010'), (5, 3, 1))  # (5, 3, 3)
        result = cpc._dipole_to_unit(
            11.32*np.ones((5, 3)), 289.59*np.ones((5, 3))
        )
        np.testing.assert_allclose(desired, result)

        desired = np.tile(cpc.igrf_dipole('2015'), (5, 2, 1))  # (5, 2, 3)
        result = cpc._dipole_to_unit(
            np.tile(np.array([-29442.0, -1501.0, 4797.1]), (5, 2, 1))
        )
        np.testing.assert_allclose(desired, result)

        desired = np.tile(cpc.igrf_dipole('2015'), (5, 2, 1))  # (5, 2, 3)
        result = cpc._dipole_to_unit(
            -29442.0*np.ones((5, 2)), -1501.0*np.ones((5, 2)),
            4797.1*np.ones((5, 2))
        )
        np.testing.assert_allclose(desired, result)

    def test_zenith_angle(self):
        """
        Compared zenith angle with NOAA
        `https://www.esrl.noaa.gov/gmd/grad/antuv/SolarCalc.jsp`_ .

        DANGER: longitude means the hour angle which is negative geographic
        longitude

        """

        zeta = cpc.zenith_angle(cp.mjd2000(2019, 8, 1), 90., 0.)
        self.assertIsNone(np.testing.assert_allclose(
            zeta, 161.79462, rtol=1e-5))

        zeta = cpc.zenith_angle(cp.mjd2000(2013, 8, 1), 77., -54.)
        self.assertIsNone(np.testing.assert_allclose(
            zeta, 117.00128, rtol=1e-5))

        zeta = cpc.zenith_angle(cp.mjd2000(2013, 3, 20, 12), 90., 0.)
        self.assertIsNone(np.testing.assert_allclose(
            zeta, 1.85573, rtol=1e-4))

        zeta = cpc.zenith_angle(cp.mjd2000(2013, 3, 20, 10), 90., 0.)
        self.assertIsNone(np.testing.assert_allclose(
            zeta, 31.85246, rtol=1e-3))

        zeta = cpc.zenith_angle(cp.mjd2000(2013, 3, 20, 14), 90., 0.)
        self.assertIsNone(np.testing.assert_allclose(
            zeta, 28.14153, rtol=1e-3))

    def test_q_response_sphere_pre7(self):

        a = 6371.2
        n = 1

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_conducting_sphere')

        C_n_mat = np.ravel(test['C_n'])
        rho_a_mat = np.ravel(test['rho_a'])
        phi_mat = np.ravel(test['phi'])
        Q_n_mat = np.ravel(test['Q_n'])

        model = np.loadtxt(
            os.path.join(ROOT, '../data/gsm_sm_coefficients',
                         'conductivity_Utada2003.dat'))

        radius = a - model[:, 0]
        # perfectly conducting core will automatically be removed
        sigma = model[:, 1]

        periods = np.logspace(np.log10(1/48), np.log10(365*24))*3600

        C_n, rho_a, phi, Q_n = cpc.q_response_1D(periods, sigma, radius, n,
                                               kind='constant')

        self.assertIsNone(np.testing.assert_allclose(C_n, C_n_mat))
        self.assertIsNone(np.testing.assert_allclose(rho_a, rho_a_mat))
        self.assertIsNone(np.testing.assert_allclose(phi, phi_mat))
        self.assertIsNone(np.testing.assert_allclose(Q_n, Q_n_mat))

    def test_q_response_sphere(self):

        a = 6371.2
        n = 1

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_conducting_sphere_thinlayer')

        C_n_mat = np.ravel(test['C_n'])
        rho_a_mat = np.ravel(test['rho_a'])
        phi_mat = np.ravel(test['phi'])
        Q_n_mat = np.ravel(test['Q_n'])

        model = np.loadtxt(cp.basicConfig['file.Earth_conductivity'])

        radius = a - model[:, 0]
        sigma = model[:, 1]

        periods = np.logspace(np.log10(1/48), np.log10(365*24))*3600

        C_n, rho_a, phi, Q_n = cpc.q_response_1D(
            periods, sigma, radius, n, kind='quadratic')

        self.assertIsNone(np.testing.assert_allclose(C_n, C_n_mat))
        self.assertIsNone(np.testing.assert_allclose(rho_a, rho_a_mat))
        self.assertIsNone(np.testing.assert_allclose(phi, phi_mat))
        self.assertIsNone(np.testing.assert_allclose(Q_n, Q_n_mat))

    def test_rotate_gauss_coeffs(self):
        """
        Compare GSM/SM rotation matrices synthezised from chaosmagpy and
        Matlab.

        """

        for reference in ['GSM', 'SM']:

            print(f'  Testing {reference} frame of reference.')

            # load spectrum to synthesize matrices in time-domain
            filepath = cp.basicConfig[f'file.{reference}_spectrum']

            try:
                data = np.load(filepath)
            except FileNotFoundError as e:
                raise ValueError(
                    f'{reference} file not found in "chaosmagpy/lib/".'
                    ' Correct reference?') from e

            data_mat = load_matfile(MATFILE_PATH, 'test_rotate_gauss_coeffs')

            time = 10*365.25*np.random.random_sample((30,))

            for freq, spec in zip(['frequency', 'frequency_ind'],
                                  ['spectrum', 'spectrum_ind']):

                matrix = cpc.synth_rotate_gauss(
                    time, data[freq], data[spec], scaled=True)

                matrix_mat = cpc.synth_rotate_gauss(
                    time, data_mat[reference + '_' + freq],
                    data_mat[reference + '_' + spec], scaled=True)

                self.assertIsNone(np.testing.assert_allclose(
                    matrix, matrix_mat, atol=1e-3))

    def test_synth_rotate_gauss(self):
        """
        Tests the accuracy of the Fourier representation of tranformation
        matrices by comparing them with directly computed matrices (they are
        considered correct).

        """

        for reference in ['gsm', 'sm']:

            print(f'  Testing {reference.upper()} frame of reference.')

            # load spectrum to synthesize matrices in time-domain
            filepath = cp.basicConfig[f'file.{reference.upper()}_spectrum']

            try:
                data = np.load(filepath)
            except FileNotFoundError as e:
                raise ValueError(
                    'Reference file "frequency_spectrum_{reference}.npz"'
                    ' not found in "chaosmagpy/lib/".'
                    ' Correct reference?') from e

            frequency = data['frequency']  # oscillations per day
            spectrum = data['spectrum']

            print("  Testing 50 times within 1996 and 2024.")
            for time in np.linspace(-4*365.25, 24*365.25, 50):

                matrix_time = cpc.synth_rotate_gauss(
                    time, frequency, spectrum, scaled=data['scaled'])

                nmax = int(np.sqrt(spectrum.shape[1] + 1) - 1)
                kmax = int(np.sqrt(spectrum.shape[2] + 1) - 1)

                if reference == 'gsm':
                    base_1, base_2, base_3 = cpc.basevectors_gsm(time)
                elif reference == 'sm':
                    base_1, base_2, base_3 = cpc.basevectors_sm(time)

                matrix = cpc.rotate_gauss(nmax, kmax, base_1, base_2, base_3)

                stat = np.amax(np.abs(matrix-np.squeeze(matrix_time)))
                print('  Computed year {:4.2f}, '
                      'max. abs. error = {:.3e}'.format(
                          time/365.25 + 2000, stat), end='')
                if stat > 0.001:
                    print(' ' + min(int(stat/0.001), 10) * '*')
                else:
                    print('')

                self.assertIsNone(np.testing.assert_allclose(
                    matrix, np.squeeze(matrix_time), rtol=1e-1, atol=1e-1))

    def test_rotate_gauss_fft(self):

        nmax = 2
        kmax = 2

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_rotate_gauss_fft')

        for reference in ['gsm', 'sm']:

            frequency, amplitude, _, _ = cpc.rotate_gauss_fft(
                nmax, kmax, step=1., N=int(365*24), filter=20,
                save_to=False, reference=reference, scaled=False)

            omega = 2*pi*frequency / (24*3600)

            # transpose and take complex conjugate (to match original output)
            omega_mat = \
                test['omega_{:}'.format(reference)].transpose((2, 1, 0))
            amplitude_mat = \
                test['amplitude_{:}'.format(reference)].transpose((2, 1, 0))
            amplitude_mat = amplitude_mat.conj()

            # absolute tolerance to account for close-to-zero matrix entries
            self.assertIsNone(np.testing.assert_allclose(
                omega_mat, omega, atol=1e-3))
            self.assertIsNone(np.testing.assert_allclose(
                amplitude_mat, amplitude, atol=1e-10))

    def test_rotate_gauss(self):

        # test 1
        time = 20  # random day
        nmax = 4
        kmax = 2

        s = timer()

        base_1, base_2, base_3 = cpc.basevectors_gsm(time)
        matrix = cpc.rotate_gauss(nmax, kmax, base_1, base_2, base_3)

        e = timer()

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_rotate_gauss')

        matrix_mat = test['m_all'].transpose()  # transposed in Matlab code
        runtime = test['runtime']

        print("  Time for matrix computation (Python): ", e - s)
        print("  Time for matrix computation (Matlab):  {:}".format(
            runtime[0, 0]))

        # absolute tolerance to account for close-to-zero matrix entries
        self.assertIsNone(np.testing.assert_allclose(
            matrix, matrix_mat, atol=1e-8))

        # test 2
        # rotate around y-axis to have dipole axis aligned with x-axis
        base_1 = np.array([0, 0, -1])
        base_2 = np.array([0, 1, 0])
        base_3 = np.array([1, 0, 0])

        nmax = 1  # only dipole terms
        kmax = 1

        matrix = cpc.rotate_gauss(nmax, kmax, base_1, base_2, base_3)
        desired = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])

        self.assertIsNone(np.testing.assert_allclose(
            matrix, desired, atol=1e-10))

        # test 3
        time = np.arange(9).reshape((3, 3))  # random day
        nmax = 4
        kmax = 2

        base_1, base_2, base_3 = cpc.basevectors_gsm(time)
        matrix = cpc.rotate_gauss(nmax, kmax, base_1, base_2, base_3)

        self.assertEqual(matrix.shape, (3, 3, 24, 8))

    def test_cartesian_to_spherical(self):

        x = -1.0
        y = 0.0
        z = 0.0

        radius = 1.0
        theta = 90.
        phi = 180.

        result = (radius, theta, phi)
        self.assertAlmostEqual(cpc.cartesian_to_spherical(x, y, z), result)

    def test_spherical_to_cartesian(self):

        radius = 1.
        theta = pi/2
        phi = pi

        x = np.array(radius) * np.sin(theta) * np.cos(phi)
        y = np.array(radius) * np.sin(theta) * np.sin(phi)
        z = np.array(radius) * np.cos(theta)

        result = (x, y, z)

        theta, phi = np.degrees(theta), np.degrees(phi)

        self.assertEqual(cpc.spherical_to_cartesian(radius, theta, phi), result)
        self.assertEqual(cpc.spherical_to_cartesian(1, theta, phi), result)
        self.assertEqual(cpc.spherical_to_cartesian(
            theta=theta, radius=radius, phi=phi), result)
        self.assertEqual(cpc.spherical_to_cartesian(
            radius=1, phi=phi, theta=theta), result)

    def test_gg_to_geo(self):

        mat = load_matfile(MATFILE_PATH, 'test_gg_to_geo')

        radius, theta, Br, Bt = cpc.gg_to_geo(mat['height'], mat['beta'],
                                              mat['X'], mat['Z'])

        np.testing.assert_allclose(radius, mat['radius'])
        np.testing.assert_allclose(theta, mat['theta'])
        np.testing.assert_allclose(Br, mat['Br'], atol=1e-6)
        np.testing.assert_allclose(Bt, mat['Bt'], atol=1e-6)

    def test_geo_to_gg(self):

        mat = load_matfile(MATFILE_PATH, 'test_geo_to_gg')

        height, beta, X, Z = cpc.geo_to_gg(mat['radius'], mat['theta'],
                                           mat['Br'], mat['Bt'])

        np.testing.assert_allclose(height, mat['height'])
        np.testing.assert_allclose(beta, mat['beta'])
        np.testing.assert_allclose(X, mat['X'], atol=1e-6)
        np.testing.assert_allclose(Z, mat['Z'], atol=1e-6)

    def test_gg_geo_gg(self):

        mat = load_matfile(MATFILE_PATH, 'test_gg_to_geo')

        radius, theta = cpc.gg_to_geo(mat['height'], mat['beta'])

        height, beta = cpc.geo_to_gg(radius, theta)

        np.testing.assert_allclose(height, mat['height'], atol=1e-10)
        np.testing.assert_allclose(beta, mat['beta'], atol=1e-10)

    def test_matrix_geo_to_base(self):

        theta_geo = np.linspace(1, 179, 10)
        phi_geo = np.linspace(-179, 179, 10)
        time = np.linspace(-300, 10000, 10)

        for reference in ['sm', 'gsm', 'mag']:
            print(f'  Testing {reference.upper()}')

            if reference == 'mag':
                base_1, base_2, base_3 = cpc.basevectors_mag()
            elif reference == 'sm':
                base_1, base_2, base_3 = cpc.basevectors_sm(time)
            elif reference == 'gsm':
                base_1, base_2, base_3 = cpc.basevectors_gsm(time)

            theta_ref, phi_ref, R = cpc.matrix_geo_to_base(
                theta_geo, phi_geo, base_1, base_2, base_3, inverse=False)

            theta_geo2, phi_geo2, R2 = cpc.matrix_geo_to_base(
                theta_ref, phi_ref, base_1, base_2, base_3, inverse=True)

            self.assertIsNone(
                np.testing.assert_allclose(theta_geo, theta_geo2))
            self.assertIsNone(
                np.testing.assert_allclose(phi_geo, phi_geo2))

            R_full = np.matmul(R, R2)
            R_full2 = np.zeros((theta_geo.size, 3, 3))
            for n in range(3):
                R_full2[..., n, n] = 1.

            self.assertIsNone(
                np.testing.assert_allclose(R_full, R_full2, atol=1e-7))

    def test_geo_to_sm(self):

        time = np.linspace(1, 100, 10)
        theta_geo = np.linspace(1, 179, 10)
        phi_geo = np.linspace(-180, 179, 10)

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_geo_to_sm')

        # reduce 2-D matrix to 1-D vectors
        theta_sm_mat = np.ravel(test['theta_sm'])
        phi_sm_mat = np.ravel(test['phi_sm'])

        theta_sm, phi_sm = cpc.transform_points(theta_geo, phi_geo,
                                              time=time, reference='sm')

        self.assertIsNone(np.testing.assert_allclose(theta_sm, theta_sm_mat))
        self.assertIsNone(np.testing.assert_allclose(phi_sm, phi_sm_mat))

    def test_geo_to_gsm(self):

        time = np.linspace(1, 100, 10)
        theta_geo = np.linspace(1, 179, 10)
        phi_geo = np.linspace(-180, 179, 10)

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_geo_to_gsm')

        # reduce 2-D matrix to 1-D vectors
        theta_gsm_mat = np.ravel(test['theta_gsm'])
        phi_gsm_mat = np.ravel(test['phi_gsm'])

        theta_gsm, phi_gsm = cpc.transform_points(
            theta_geo, phi_geo, time=time, reference='gsm')

        self.assertIsNone(np.testing.assert_allclose(theta_gsm, theta_gsm_mat))
        self.assertIsNone(np.testing.assert_allclose(phi_gsm, phi_gsm_mat))

    def test_geo_to_base(self):

        time = np.linspace(1, 100, 10)
        theta_geo = np.linspace(1, 179, 10)
        phi_geo = np.linspace(-180, 179, 10)

        # TEST GSM COORDINATES
        # GSM test: load matfile
        test = load_matfile(MATFILE_PATH, 'test_geo_to_gsm')

        # reduce 2-D matrix to 1-D vectors
        theta_gsm_mat = np.ravel(test['theta_gsm'])
        phi_gsm_mat = np.ravel(test['phi_gsm'])

        gsm_1, gsm_2, gsm_3 = cpc.basevectors_gsm(time)

        theta_gsm, phi_gsm = cpc.geo_to_base(
            theta_geo, phi_geo, gsm_1, gsm_2, gsm_3)

        self.assertIsNone(np.testing.assert_allclose(theta_gsm, theta_gsm_mat))
        self.assertIsNone(np.testing.assert_allclose(phi_gsm, phi_gsm_mat))

        # test the inverse option: GEO -> GSM -> GEO
        theta_geo2, phi_geo2 = cpc.geo_to_base(
            theta_gsm, phi_gsm, gsm_1, gsm_2, gsm_3, inverse=True)

        self.assertIsNone(np.testing.assert_allclose(theta_geo, theta_geo2))
        self.assertIsNone(np.testing.assert_allclose(
            phi_geo, cpc.center_azimuth(phi_geo2)))

        # test the inverse option: GSM -> GEO -> GSM
        theta_gsm2, phi_gsm2 = cpc.geo_to_base(
            theta_geo2, phi_geo2, gsm_1, gsm_2, gsm_3)

        self.assertIsNone(np.testing.assert_allclose(theta_gsm, theta_gsm2))
        self.assertIsNone(np.testing.assert_allclose(
            cpc.center_azimuth(phi_gsm), cpc.center_azimuth(phi_gsm2)))

        # TEST SM COORDINATES
        # SM test: load matfile
        test = load_matfile(MATFILE_PATH, 'test_geo_to_sm')

        # reduce 2-D matrix to 1-D vectors
        theta_sm_mat = np.ravel(test['theta_sm'])
        phi_sm_mat = np.ravel(test['phi_sm'])

        sm_1, sm_2, sm_3 = cpc.basevectors_sm(time)

        theta_sm, phi_sm = cpc.geo_to_base(
            theta_geo, phi_geo, sm_1, sm_2, sm_3)

        self.assertIsNone(np.testing.assert_allclose(theta_sm, theta_sm_mat))
        self.assertIsNone(np.testing.assert_allclose(phi_sm, phi_sm_mat))

        # test the inverse option: GEO -> SM -> GEO
        theta_geo2, phi_geo2 = cpc.geo_to_base(
            theta_sm, phi_sm, sm_1, sm_2, sm_3, inverse=True)

        self.assertIsNone(np.testing.assert_allclose(theta_geo, theta_geo2))
        self.assertIsNone(np.testing.assert_allclose(phi_geo, phi_geo2))

        # test the inverse option: SM -> GEO -> SM
        theta_sm2, phi_sm2 = cpc.geo_to_base(
            theta_geo2, phi_geo2, sm_1, sm_2, sm_3)

        self.assertIsNone(np.testing.assert_allclose(theta_sm, theta_sm2))
        self.assertIsNone(np.testing.assert_allclose(phi_sm, phi_sm2))

    def test_basevectors_use(self):

        # test local transformation at specific point
        R = np.stack(cpc.basevectors_use(
            theta=90., phi=180.), axis=-1)
        desired = np.array([[-1, 0, 0], [0, 0, -1], [0, -1, 0]])

        self.assertIsNone(np.testing.assert_allclose(R, desired, atol=1e-10))

        # test local transformation at specific point
        R = np.stack(cpc.basevectors_use(
            theta=90., phi=-90.), axis=-1)
        desired = np.array([[0, 0, 1], [-1, 0, 0], [0, -1, 0]])

        self.assertIsNone(np.testing.assert_allclose(R, desired, atol=1e-10))

    def test_sun_position(self):

        start = cp.mjd2000(1910, 1, 1)
        end = cp.mjd2000(2090, 12, 31)
        time = np.linspace(start, end, 100)

        theta, phi = cpc.sun_position(time)

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_sun_position')

        # reduce 2-D matrix to 1-D vectors
        theta_mat = np.ravel(test['theta'])
        phi_mat = np.ravel(test['phi'])

        self.assertIsNone(np.testing.assert_allclose(theta, theta_mat))
        self.assertIsNone(np.testing.assert_allclose(phi, phi_mat))

        # test position of sun close to zenith in Greenwich
        # (lat 51.48, lon -0.0077) on July 18, 2018, 12h06 (from
        # sunearthtools.com):

        time_gmt = cp.mjd2000(2018, 7, 18, 12, 6)
        lat = 51.4825766  # geogr. latitude
        lon = -0.0076589  # geogr. longitude
        el = 59.59  # elevation above horizon
        az = 179.86  # angle measured from north (sun close to zenith)

        theta_gmt = 180. - (el + lat)  # subsolar colat. given geogr. lat.
        phi_gmt = 180. + (lon - az)  # subsolar phi given geogr. longitude

        theta, phi = cpc.sun_position(time_gmt)

        self.assertAlmostEqual(theta_gmt, theta, places=0)
        self.assertAlmostEqual(phi_gmt, phi, places=0)

    def test_center_azimuth(self):

        angle = 180.

        self.assertAlmostEqual(cpc.center_azimuth(0.25*angle), 0.25*angle)
        self.assertAlmostEqual(cpc.center_azimuth(0.75*angle), 0.75*angle)
        self.assertAlmostEqual(cpc.center_azimuth(1.25*angle), -0.75*angle)
        self.assertAlmostEqual(cpc.center_azimuth(1.75*angle), -0.25*angle)
        self.assertAlmostEqual(cpc.center_azimuth(-0.25*angle), -0.25*angle)
        self.assertAlmostEqual(cpc.center_azimuth(-0.75*angle), -0.75*angle)
        self.assertAlmostEqual(cpc.center_azimuth(-1.25*angle), 0.75*angle)
        self.assertAlmostEqual(cpc.center_azimuth(-1.75*angle), 0.25*angle)


if __name__ == '__main__':
    main()



# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

import numpy as np
import os
import textwrap
from unittest import TestCase, main
from datetime import datetime, timedelta
from chaosmagpy import data_utils as cpd

ROOT = os.path.abspath(os.path.dirname(__file__))
MATFILE_PATH = os.path.join(ROOT, 'data/CHAOS_test.mat')

# check if mat-file exists in tests directory
if os.path.isfile(MATFILE_PATH) is False:
    MATFILE_PATH = str(input('Matfile path for data_utils test?: '))


class DataUtils(TestCase):
    def setUp(self):

        print(textwrap.dedent(f"""\

            {"":-^70}
            Running {self._testMethodName}:
            """))

    def test_load_RC_datfile(self):

        # make sure link is not broken
        cpd.load_RC_datfile(filepath=None, parse_dates=None)

    def test_time_conversion(self):

        size = (10,)
        days = np.random.randint(365., size=size)
        hours = np.random.randint(24, size=size)
        minutes = np.random.randint(60, size=size)
        seconds = np.random.randint(60, size=size)

        for day, hour, minute, second in zip(days, hours, minutes, seconds):
            date = (timedelta(days=int(day)) +
                    datetime(1990, 1, 1, hour, minute, second))

            mjd = cpd.mjd2000(date.year, date.month, date.day, date.hour,
                              date.minute, date.second)

            dyear = cpd.mjd_to_dyear(mjd, leap_year=True)
            mjd2 = cpd.dyear_to_mjd(dyear, leap_year=True)
            self.assertIsNone(np.testing.assert_allclose(mjd2, mjd, atol=1e-8))

            dyear = cpd.mjd_to_dyear(mjd, leap_year=False)
            mjd2 = cpd.dyear_to_mjd(dyear, leap_year=False)
            self.assertIsNone(np.testing.assert_allclose(mjd2, mjd, atol=1e-8))

            dyear = cpd.mjd_to_dyear(mjd, leap_year=True)
            mjd2 = cpd.dyear_to_mjd(dyear, leap_year=False)
            self.assertRaises(
                AssertionError, lambda: np.testing.assert_allclose(
                    mjd2, mjd, atol=1e-8))

            dyear = cpd.mjd_to_dyear(mjd, leap_year=False)
            mjd2 = cpd.dyear_to_mjd(dyear, leap_year=True)
            self.assertRaises(
                AssertionError, lambda: np.testing.assert_allclose(
                    mjd2, mjd, atol=1e-8))

            dyear = cpd.mjd_to_dyear(mjd, leap_year=False)
            mjd2 = cpd.dyear_to_mjd(dyear, leap_year=None)
            self.assertRaises(
                AssertionError, lambda: np.testing.assert_allclose(
                    mjd2, mjd, atol=1e-8))

    def test_mjd2000_broadcasting(self):
        """
        Ensure broadcasting works
        """

        actual = cpd.mjd2000(np.arange(2000, 2003), 2, 1)
        desired = np.array([  31.,  397.,  762.])
        np.testing.assert_allclose(actual, desired)

        actual = cpd.mjd2000(2000, 2, 1, np.arange(3))
        desired = 31. + np.arange(3)/24
        np.testing.assert_allclose(actual, desired)

    def test_mjd_to_dyear(self):

        self.assertEqual(cpd.mjd_to_dyear(0.0), 2000.0)
        self.assertEqual(cpd.mjd_to_dyear(366.0), 2001.0)
        self.assertEqual(cpd.mjd_to_dyear(731.0), 2002.0)
        self.assertEqual(cpd.mjd_to_dyear(1096.0), 2003.0)
        self.assertEqual(cpd.mjd_to_dyear(4*365.25), 2004.0)

        self.assertEqual(cpd.mjd_to_dyear(0.0, leap_year=False), 2000.0)
        self.assertEqual(cpd.mjd_to_dyear(365.25, leap_year=False), 2001.0)
        self.assertEqual(cpd.mjd_to_dyear(2*365.25, leap_year=False), 2002.0)
        self.assertEqual(cpd.mjd_to_dyear(3*365.25, leap_year=False), 2003.0)
        self.assertEqual(cpd.mjd_to_dyear(4*365.25, leap_year=False), 2004.0)


if __name__ == '__main__':
    main()



# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

"""
ChaosMagPy is a simple Python package for evaluating the CHAOS geomagnetic
field model and other models of Earth's magnetic field. The latest CHAOS model
is available at http://www.spacecenter.dk/files/magnetic-models/CHAOS-7/.

>>> import chaosmagpy as cp
>>> chaos = cp.load_CHAOS_matfile('CHAOS-7.mat')

The following modules are available:

* chaosmagpy.chaos (functions for loading well-known field models)
* chaosmagpy.coordinate_utils (tools for transforming coordinates/components)
* chaosmagpy.model_utils (tools for evaluating field models)
* chaosmagpy.plot_utils (tools for plotting)
* chaosmagpy.data_utils (tools for reading/writing data files)
* chaosmagpy.config_utils (tools for customizing ChaosMagPy parameters)

More information on how to use ChaosMagPy can be found in the documentation
available at https://chaosmagpy.readthedocs.io/en/.

"""

__all__ = [
    "CHAOS", "load_CHAOS_matfile", "load_CHAOS_shcfile",
    "load_CovObs_txtfile", "load_gufm1_txtfile", "load_CALS7K_txtfile",
    "load_IGRF_txtfile", "basicConfig", "synth_values", "mjd2000",
    "timestamp", "dyear_to_mjd", "mjd_to_dyear"
]

__version__ = "0.13.1"

from .chaos import (
    CHAOS,
    load_CHAOS_matfile,
    load_CHAOS_shcfile,
    load_CovObs_txtfile,
    load_gufm1_txtfile,
    load_CALS7K_txtfile,
    load_IGRF_txtfile
)

from .config_utils import basicConfig
from .model_utils import synth_values
from .data_utils import mjd2000, timestamp, dyear_to_mjd, mjd_to_dyear



# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

import numpy as np
import os
import textwrap
from unittest import TestCase, main
from chaosmagpy import model_utils as m
from timeit import default_timer as timer
from math import pi

try:
    from tests.helpers import load_matfile
except ImportError:
    from helpers import load_matfile


RAD = pi / 180
R_REF = 6371.2  # reference radius in kilometers

ROOT = os.path.abspath(os.path.dirname(__file__))
MATFILE_PATH = os.path.join(ROOT, 'data/CHAOS_test.mat')

# check if mat-file exists in tests directory
if os.path.isfile(MATFILE_PATH) is False:
    MATFILE_PATH = str(input('Matfile path for coordinate_utils test?: '))


class ModelUtils(TestCase):
    def setUp(self):

        print(textwrap.dedent(f"""\

            {"":-^70}
            Running {self._testMethodName}:
            """))

    def test_design_gauss(self):

        radius = 6371*np.ones((15,))
        theta = np.arange(15)
        phi = np.arange(15)

        # order equal to zero
        A = np.array(m.design_gauss(radius, theta, phi, nmax=3))
        A2 = np.array(m.design_gauss(radius, theta, phi, nmax=3, mmax=0))

        index = [0, 3, 8]
        np.testing.assert_equal(A[:, :, index], A2)

        # order equal or less than 1
        A = np.array(m.design_gauss(radius, theta, phi, nmax=3))
        A2 = np.array(m.design_gauss(radius, theta, phi, nmax=3, mmax=1))

        index = [0, 1, 2, 3, 4, 5, 8, 9, 10]
        np.testing.assert_equal(A[:, :, index], A2)

    def test_design_gauss_multidim(self):

        radius = 6371.
        theta = np.arange(30).reshape((3, 1, 10))
        phi = np.arange(30).reshape((3, 10, 1))
        coeffs = np.arange(8) + 1.

        Br, Bt, Bp = m.synth_values(coeffs, radius, theta, phi, nmax=2)
        Ar, At, Ap = m.design_gauss(radius, theta, phi, nmax=2)

        np.testing.assert_allclose(Br, Ar@coeffs)
        np.testing.assert_allclose(Bt, At@coeffs)
        np.testing.assert_allclose(Bp, Ap@coeffs)

    def test_degree_correlation(self):

        nmax = 4
        coeffs = np.random.random((int(nmax*(nmax+2)),))

        np.testing.assert_equal(
            m.degree_correlation(coeffs, coeffs), np.ones((nmax,)))

    def test_power_spectrum(self):

        shape = (15, 24, 3)  # nmax = (3, 4, 1)
        coeffs = np.ones(shape)

        R_n = m.power_spectrum(coeffs, source='internal')  # axis=-1
        R_n_desired = 6. * np.ones((15, 24, 1))
        np.testing.assert_equal(R_n, R_n_desired)

        R_n = m.power_spectrum(coeffs, source='internal', axis=0)
        R_n_desired = np.array([6., 15., 28.])[:, None, None] * np.ones(
            (3, 24, 3))
        np.testing.assert_equal(R_n, R_n_desired)

        R_n = m.power_spectrum(coeffs, source='external', axis=1)
        R_n_desired = np.array([3., 10., 21., 36])[None, :, None] * np.ones(
            (15, 4, 3))
        np.testing.assert_equal(R_n, R_n_desired)

        R_n = m.power_spectrum(coeffs, source='toroidal')
        R_n_desired = 2. * np.ones((15, 24, 1))
        np.testing.assert_equal(R_n, R_n_desired)

    def test_synth_values_mmax(self):

        theta = 79.
        phi = 13.
        radius = R_REF

        # function for quick testing with "true" solution
        np.testing.assert_allclose(
            m.synth_values([1., 2., 3., 4., 5., 6., 7., 8.],
                           radius, theta, phi),
            m.synth_values([1., 2., 3., 4., 5., 6., 7., 8.],
                           radius, theta, phi, nmax=None, mmax=None))

        np.testing.assert_allclose(
            m.synth_values([1., 2., 3., 0., 0., 0., 0., 0.],
                           radius, theta, phi),
            m.synth_values([1., 2., 3., 4., 5., 6., 7., 8.],
                           radius, theta, phi, nmax=1, mmax=None))

        np.testing.assert_allclose(
            m.synth_values([1., 2., 3., 4., 5., 6., 0., 0.],
                           radius, theta, phi),
            m.synth_values([1., 2., 3., 4., 5., 6.],
                           radius, theta, phi, nmax=2, mmax=1))

        np.testing.assert_allclose(
            m.synth_values([1., 0., 0., 0., 0., 0., 0., 0.],
                           radius, theta, phi),
            m.synth_values([1.],
                           radius, theta, phi, nmax=1, mmax=0))

        np.testing.assert_allclose(
            m.synth_values([1., 2., 3., 4., 5., 6., 0., 0.],
                           radius, theta, phi),
            m.synth_values([1., 2., 3., 4., 5., 6.],
                           radius, theta, phi, nmax=None, mmax=1))

    def test_synth_values_nmin(self):

        theta = 79.
        phi = 13.
        radius = R_REF

        theta_ls = theta*np.ones((15,))
        phi_ls = phi*np.ones((27,))
        phi_grid, theta_grid = np.meshgrid(phi_ls, theta_ls)  # 2-D
        radius_grid = R_REF * np.ones(phi_grid.shape)  # 2-D
        coeffs_grid = np.random.random()*np.ones(phi_grid.shape + (24,))  # 3-D

        coeffs_grid_min = np.copy(coeffs_grid)
        coeffs_grid_min[..., :8] = 0.  # nmax=4

        # exhaustive inputs
        B = m.synth_values(coeffs_grid_min, radius_grid, theta_grid, phi_grid)

        # function for quick testing with "true" solution
        def test(field):
            self.assertIsNone(np.testing.assert_allclose(B, field))

        test(m.synth_values(coeffs_grid[..., 8:], radius, theta_grid, phi_grid,
                            nmin=3, nmax=4))
        test(m.synth_values(coeffs_grid_min[..., 8:], radius, theta_grid,
                            phi_grid, nmin=3, nmax=4))

    def test_synth_values_inputs(self):

        theta = 79.
        phi = 13.
        radius = R_REF

        theta_ls = theta*np.ones((15,))
        phi_ls = phi*np.ones((27,))
        phi_grid, theta_grid = np.meshgrid(phi_ls, theta_ls)  # 2-D
        radius_grid = R_REF * np.ones(phi_grid.shape)  # 2-D
        coeffs_grid = np.random.random()*np.ones(phi_grid.shape + (24,))  # 3-D

        # exhaustive inputs
        B = m.synth_values(coeffs_grid, radius_grid, theta_grid, phi_grid)

        # function for quick testing with "true" solution
        def test(field):
            self.assertIsNone(np.testing.assert_allclose(B, field))

        # one input () dimension: float
        test(m.synth_values(coeffs_grid, radius, theta_grid, phi_grid))
        test(m.synth_values(coeffs_grid, radius_grid, theta, phi_grid))
        test(m.synth_values(coeffs_grid, radius_grid, theta_grid, phi))

        # one input (1,) dimension: one element array
        test(m.synth_values(
            coeffs_grid[0, 0], radius_grid, theta_grid, phi_grid))
        test(m.synth_values(
            coeffs_grid, radius_grid[0, 0], theta_grid, phi_grid))
        test(m.synth_values(
            coeffs_grid, radius_grid, theta_grid[0, 0], phi_grid))
        test(m.synth_values(
            coeffs_grid, radius_grid, theta_grid, phi_grid[0, 0]))

        # GRID MODE
        test(m.synth_values(
            coeffs_grid, radius_grid, theta_ls, phi_ls, grid=True))
        test(m.synth_values(coeffs_grid, radius, theta_ls, phi_ls, grid=True))
        test(m.synth_values(
            coeffs_grid[0, 0], radius_grid, theta_ls, phi_ls, grid=True))
        test(m.synth_values(
            coeffs_grid[0, 0], radius_grid[0, 0], theta_ls, phi_ls, grid=True))

        # test multi-dimensional grid

        # full grid
        coeffs_grid = np.random.random((24,))
        radius_grid = radius * np.ones((2, 5, 6, 4))
        phi_grid = phi * np.ones((2, 5, 6, 4))
        theta_grid = theta * np.ones((2, 5, 6, 4))

        B = m.synth_values(coeffs_grid, radius_grid, theta_grid, phi_grid)

        # reduced grid
        radius_grid2 = radius * np.ones((2, 1, 6, 4))
        phi_grid2 = phi * np.ones((2, 5, 1, 1))
        theta_grid2 = theta * np.ones((1, 5, 6, 4))

        test(m.synth_values(coeffs_grid, radius_grid2, theta_grid2, phi_grid2))

    def test_synth_values_grid(self):

        radius = R_REF
        theta = np.linspace(0., 180., num=4000)
        phi = np.linspace(-180., 180., num=3000)
        phi_grid, theta_grid = np.meshgrid(phi, theta)
        coeffs = np.random.random((24,))

        s = timer()
        B_grid = m.synth_values(
            coeffs, radius, theta_grid, phi_grid, grid=False)
        e = timer()
        print("  Time for 'grid=False' computation: ", e - s)

        s = timer()
        B_grid2 = m.synth_values(
            coeffs, radius, theta[..., None], phi[None, ...], grid=False)
        e = timer()
        print("  Time for broadcasted 'grid=False' computation: ", e - s)

        s = timer()
        B = m.synth_values(
            coeffs, radius, theta, phi, grid=True)
        e = timer()
        print("  Time for 'grid=True' computation: ", e - s)

        for comp, comp_grid in zip(B_grid, B_grid2):
            np.testing.assert_allclose(comp, comp_grid)

        for comp, comp_grid in zip(B_grid2, B):
            np.testing.assert_allclose(comp, comp_grid)

    def test_synth_values_poles(self):

        radius = 6371.2
        theta = np.array([0., 180.])
        phi = np.linspace(0., 360., num=10, endpoint=False)

        # dipole aligned with x-axis, B = -grad(V)
        # North pole: Bx = -1, Bt = -1
        # South pole: Bx = -1, Bt = 1
        Br, Bt, Bp = m.synth_values([0., 1., 0], radius, theta, phi, grid=True)

        # north pole (theta, phi) = (0., 0.)
        npole = (0, 0)
        np.testing.assert_allclose(Br[npole], 0.)
        np.testing.assert_allclose(Bt[npole], -1.)
        np.testing.assert_allclose(Bp[npole], 0.)

        Bx = (np.cos(np.radians(0.))*np.cos(np.radians(phi))*Bt[0, :]
              - np.sin(np.radians(phi))*Bp[0, :])
        np.testing.assert_allclose(Bx, -np.ones((10,)))

        # south pole (theta, phi) = (180., 0.)
        spole = (1, 0)
        np.testing.assert_allclose(Br[spole], 0.)
        np.testing.assert_allclose(Bt[spole], 1.)
        np.testing.assert_allclose(Bp[spole], 0.)

        Bx = (np.cos(np.radians(180.))*np.cos(np.radians(phi))*Bt[1, :]
              - np.sin(np.radians(phi))*Bp[1, :])
        np.testing.assert_allclose(Bx, -np.ones((10,)))

    def test_design_matrix(self):
        """
        Test matrices for time-dependent field model using B-spline basis.

        """

        n_data = int(300)
        t_start = 1997.1
        t_end = 2018.1
        n_breaks = int((t_end - t_start) / 0.5 + 1)
        time = np.linspace(t_start, t_end, num=n_data)
        radius = R_REF * np.ones(time.shape)
        theta = np.linspace(1, 179, num=n_data)
        phi = np.linspace(-180, 179, num=n_data)
        n_static = int(80)
        n_tdep = int(20)
        order = int(6)  # order of spline basis functions (4 = cubic)

        # create a knot vector without endpoint repeats and # add endpoint
        # repeats as appropriate for spline degree p
        knots = np.linspace(t_start, t_end, num=n_breaks)
        knots = m.augment_breaks(knots, order)

        s = timer()
        G_radius, G_theta, G_phi = m.design_matrix(knots, order, n_tdep, time,
                                                   radius, theta, phi,
                                                   n_static=n_static)
        e = timer()

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_design_matrix')

        G_radius_mat = test['G_radius']
        G_theta_mat = test['G_theta']
        G_phi_mat = test['G_phi']
        runtime = test['runtime']

        print("  Time for design_matrix computation (Python): ", e - s)
        print("  Time for design_matrix computation (Matlab):  {:}".format(
            runtime[0, 0]))

        self.assertIsNone(
            np.testing.assert_allclose(G_radius, G_radius_mat, atol=1e-5))
        self.assertIsNone(
            np.testing.assert_allclose(G_theta, G_theta_mat, atol=1e-5))
        self.assertIsNone(
            np.testing.assert_allclose(G_phi, G_phi_mat, atol=1e-5))

    def test_colloc_matrix(self):

        breaks = np.linspace(0., 300., 20)
        order = 6
        knots = m.augment_breaks(breaks, order)

        x = np.linspace(0., 300., 1000)

        test = load_matfile(MATFILE_PATH, 'test_colloc_matrix')

        colloc_m = test['colloc']

        for deriv in range(order):
            print(f"Checking deriv = {deriv} of collocation matrix.")

            colloc = m.colloc_matrix(x, knots, order, deriv=deriv)

            self.assertIsNone(np.testing.assert_allclose(
                colloc, colloc_m[deriv::order], atol=1e-5))


if __name__ == '__main__':
    main()



# Copyright (C) 2024 Clemens Kloss
#
# This file is part of ChaosMagPy.
#
# ChaosMagPy is released under the MIT license. See LICENSE in the root of the
# repository for full licensing details.

import os
import numpy as np
import textwrap
import chaosmagpy as cp
import matplotlib.pyplot as plt
from unittest import TestCase, main

try:
    from tests.helpers import load_matfile
except ImportError:
    from helpers import load_matfile

R_REF = 6371.2  # reference radius in km
ROOT = os.path.abspath(os.path.dirname(__file__))
MATFILE_PATH = os.path.join(ROOT, 'data/CHAOS_test.mat')
CHAOS_PATH = os.path.join(ROOT, 'data/CHAOS-6-x7.mat')
CHAOS_PATH_SHC = os.path.join(ROOT, 'data/CHAOS-6-x7_tdep.shc')

# check if mat-file exists in tests directory
if os.path.isfile(MATFILE_PATH) is False:
    MATFILE_PATH = str(input('Matfile path for chaosmagpy test?: '))


class Chaos(TestCase):
    def setUp(self):

        print(textwrap.dedent(f"""\

            {"":-^70}
            Running {self._testMethodName}:
            """))

    def test_synth_euler_angles(self):

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        test = load_matfile(MATFILE_PATH, 'test_synth_euler_angles')
        time = np.squeeze(test['time'])

        swarm_c = model.synth_euler_angles(time, 'swarm_c')

        self.assertIsNone(np.testing.assert_allclose(
            swarm_c, test['swarm_c']))

    def test_load_CHAOS(self):

        cp.load_CHAOS_matfile(CHAOS_PATH)
        cp.load_CHAOS_shcfile(CHAOS_PATH_SHC)
        cp.chaos.BaseModel.from_shc(CHAOS_PATH_SHC)

    def test_save_matfile(self):

        seq = np.random.randint(0, 10, size=(5,))
        filename = 'CHAOS-tmp_' + ''.join([str(a) for a in seq]) + '.mat'
        filepath = os.path.join(ROOT, filename)

        model = cp.load_CHAOS_matfile(CHAOS_PATH)
        print('  ', end='')  # indent line by two withespaces
        model.save_matfile(filepath)

        def test(x, y):
            if isinstance(x, str) or isinstance(y, str):
                # convert unicode to str
                np.testing.assert_string_equal(str(x), str(y))
            else:
                np.testing.assert_allclose(x, y, atol=1e-10)

        chaos = cp.data_utils.load_matfile(CHAOS_PATH, variable_names=[
            'pp', 'model_ext', 'model_Euler', 'g'])
        chaos_out = cp.data_utils.load_matfile(filepath, variable_names=[
            'pp', 'model_ext', 'model_Euler', 'g'])

        pp = chaos['pp']
        pp_out = chaos_out['pp']

        for key in ['order', 'dim', 'pieces', 'form', 'coefs', 'breaks']:
            test(pp[key], pp_out[key])

        test(chaos['g'], chaos_out['g'])

        model_ext = chaos['model_ext']
        model_ext_out = chaos_out['model_ext']

        for key in ['m_Dst', 'm_gsm', 'm_sm', 'q10', 'qs11', 't_break_q10',
                    't_break_qs11']:
            test(model_ext[key], model_ext_out[key])

        model_Euler = chaos['model_Euler']
        model_Euler_out = chaos_out['model_Euler']

        for key in ['alpha', 'beta', 'gamma', 't_break_Euler']:
            var = model_Euler[key]
            var_out = model_Euler_out[key]
            test(var.shape, var_out.shape)

            for value, value_out in zip(np.ravel(var), np.ravel(var_out)):
                test(value, value_out)

        print(f"  Removing file {filepath}")
        os.remove(filepath)

    def test_save_shcfile(self):

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        filepath = os.path.join(ROOT, 'CHAOS_tdep.shc')
        model.save_shcfile(filepath, model='tdep')
        model2 = cp.load_CHAOS_shcfile(filepath)

        coeffs_tdep_mat = model.model_tdep.coeffs
        coeffs_tdep_shc = model2.model_tdep.coeffs

        np.testing.assert_allclose(coeffs_tdep_shc, coeffs_tdep_mat, atol=1e-8)

        print(f"Removing file {filepath}")
        os.remove(filepath)

        # static internal field

        filepath = os.path.join(ROOT, 'CHAOS_static.shc')
        model.save_shcfile(filepath, model='static')
        model2 = cp.load_CHAOS_shcfile(filepath)

        coeffs_stat_mat = model.model_static.coeffs
        coeffs_stat_shc = model2.model_static.coeffs

        np.testing.assert_allclose(coeffs_stat_shc, coeffs_stat_mat, atol=1e-8)

        print(f"Removing file {filepath}")
        os.remove(filepath)

    def test_io_shcfile(self):

        # piecewise constant, one piece

        order = 1
        breaks = [0., 10.]
        knots = cp.model_utils.augment_breaks(breaks, order)

        model = cp.chaos.BaseModel.from_bspline(
            name='BaseModel',
            knots=knots,
            coeffs=np.random.uniform(-100., 100., 3).reshape((1, 3)),
            order=order
        )

        filepath = os.path.join(ROOT, 'BaseModel_order1_piece1.shc')
        model.to_shc(filepath)
        model2 = cp.chaos.BaseModel.from_shc(filepath)

        np.testing.assert_allclose(model.coeffs, model2.coeffs)

        print(f"Removing file {filepath}")
        os.remove(filepath)

        # piecewise constant, multiple pieces

        order = 1
        breaks = [0., 10., 30., 52., 79., 100.]  # 5 pieces
        knots = cp.model_utils.augment_breaks(breaks, order)

        model = cp.chaos.BaseModel.from_bspline(
            name='BaseModel',
            knots=knots,
            coeffs=np.random.uniform(-100., 100., 15).reshape((5, 3)),
            order=order
        )

        filepath = os.path.join(ROOT, 'BaseModel_order1_piece5.shc')
        model.to_shc(filepath)
        model2 = cp.chaos.BaseModel.from_shc(filepath)

        np.testing.assert_allclose(model.coeffs, model2.coeffs)

        print(f"Removing file {filepath}")
        os.remove(filepath)

        # piecewise linear, one piece

        order = 2
        breaks = [0., 10.]
        knots = cp.model_utils.augment_breaks(breaks, order)

        model = cp.chaos.BaseModel.from_bspline(
            name='BaseModel',
            knots=knots,
            coeffs=np.random.uniform(-100., 100., 6).reshape((2, 3)),
            order=order
        )

        filepath = os.path.join(ROOT, 'BaseModel_order2_piece1.shc')
        model.to_shc(filepath)
        model2 = cp.chaos.BaseModel.from_shc(filepath)

        np.testing.assert_allclose(model.coeffs, model2.coeffs)

        print(f"Removing file {filepath}")
        os.remove(filepath)

        # piecewise linear, multiple pieces

        order = 2
        breaks = [0., 10., 30., 52., 79., 100.]
        knots = cp.model_utils.augment_breaks(breaks, order)

        model = cp.chaos.BaseModel.from_bspline(
            name='BaseModel',
            knots=knots,
            coeffs=np.random.uniform(-100., 100., 18).reshape((6, 3)),
            order=order
        )

        filepath = os.path.join(ROOT, 'BaseModel_order2_piece5.shc')
        model.to_shc(filepath)
        model2 = cp.chaos.BaseModel.from_shc(filepath)

        np.testing.assert_allclose(model.coeffs, model2.coeffs, atol=2e-5)

        print(f"Removing file {filepath}")
        os.remove(filepath)

        # higher order B-spline, one pieces

        order = 3
        breaks = [0., 10.]
        knots = cp.model_utils.augment_breaks(breaks, order)

        model = cp.chaos.BaseModel.from_bspline(
            name='BaseModel',
            knots=knots,
            coeffs=np.random.uniform(-100., 100., 9).reshape((3, 3)),
            order=order
        )

        filepath = os.path.join(ROOT, 'BaseModel_order3_piece1.shc')
        model.to_shc(filepath)
        model2 = cp.chaos.BaseModel.from_shc(filepath)

        np.testing.assert_allclose(model.coeffs, model2.coeffs, atol=2e-5)

        print(f"Removing file {filepath}")
        os.remove(filepath)

        # higher order B-spline, multiple pieces

        order = 3
        breaks = [0., 10., 30., 52., 79., 100.]
        knots = cp.model_utils.augment_breaks(breaks, order)

        model = cp.chaos.BaseModel.from_bspline(
            name='BaseModel',
            knots=knots,
            coeffs=np.random.uniform(-100., 100., 21).reshape((7, 3)),
            order=order
        )

        filepath = os.path.join(ROOT, 'BaseModel_order3_piece5.shc')
        model.to_shc(filepath)
        model2 = cp.chaos.BaseModel.from_shc(filepath)

        np.testing.assert_allclose(model.coeffs, model2.coeffs, atol=2e-5)

        print(f"Removing file {filepath}")
        os.remove(filepath)

    def test_complete_forward(self):

        n_data = int(300)
        t_start = -200.0
        t_end = 6000.0
        time = np.linspace(t_start, t_end, num=n_data)
        radius = R_REF * np.ones(time.shape)
        theta = np.linspace(1, 179, num=n_data)
        phi = np.linspace(-180, 179, num=n_data)

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        B_radius, B_theta, B_phi = model(time, radius, theta, phi)

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_complete_forward')

        B_radius_mat = np.ravel(test['B_radius'])
        B_theta_mat = np.ravel(test['B_theta'])
        B_phi_mat = np.ravel(test['B_phi'])

        for component in ['B_radius', 'B_theta', 'B_phi']:
            res = np.abs(eval(component) - eval('_'.join((component, 'mat'))))
            print('  -------------------')
            print(f'  {component}:')
            print('  MAE =', np.mean(res), 'nT')
            print('  RMSE =', np.sqrt(np.mean(res**2)), 'nT')
            print('  Max Error =', np.amax(res), 'nT')
            print('  Min Error =', np.amin(res), 'nT')

        np.testing.assert_allclose(
            B_radius, B_radius_mat, rtol=1e-7, atol=1e-2)
        np.testing.assert_allclose(
            B_theta, B_theta_mat, rtol=1e-7, atol=1e-2)
        np.testing.assert_allclose(
            B_phi, B_phi_mat, rtol=1e-7, atol=1e-2)

    def test_surface_field(self):

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        time = 2015.
        radius = R_REF
        theta = np.linspace(1, 179, num=180)
        phi = np.linspace(-180, 179, num=360)

        B_radius, B_theta, B_phi = model.synth_values_tdep(
            time, radius, theta, phi, nmax=13, grid=True, deriv=0)

        B_radius = np.ravel(B_radius, order='F')  # ravel column-major
        B_theta = np.ravel(B_theta, order='F')
        B_phi = np.ravel(B_phi, order='F')

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_surface_field')

        B_radius_mat = np.ravel(test['B_radius'])
        B_theta_mat = np.ravel(test['B_theta'])
        B_phi_mat = np.ravel(test['B_phi'])

        for component in ['B_radius', 'B_theta', 'B_phi']:
            res = np.abs(eval(component) - eval('_'.join((component, 'mat'))))
            print('  -------------------')
            print(f'  {component}:')
            print('  MAE =', np.mean(res), 'nT')
            print('  RMSE =', np.sqrt(np.mean(res**2)), 'nT')
            print('  Max Error =', np.amax(res), 'nT')
            print('  Min Error =', np.amin(res), 'nT')

        self.assertIsNone(np.testing.assert_allclose(
            B_radius, B_radius_mat))
        self.assertIsNone(np.testing.assert_allclose(
            B_theta, B_theta_mat))
        self.assertIsNone(np.testing.assert_allclose(
            B_phi, B_phi_mat))

    def test_sv_timeseries(self):

        # some observatory location
        radius = R_REF
        theta = 90-14.308
        phi = -16.950

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        time = np.linspace(model.model_tdep.breaks[0],
                           model.model_tdep.breaks[-1], num=1000)

        B_radius, B_theta, B_phi = model.synth_values_tdep(
            time, radius, theta, phi, nmax=16, deriv=1)

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_sv_timeseries')

        B_radius_mat = np.ravel(test['B_radius'])
        B_theta_mat = np.ravel(test['B_theta'])
        B_phi_mat = np.ravel(test['B_phi'])

        for component in ['B_radius', 'B_theta', 'B_phi']:
            res = np.abs(eval(component) - eval('_'.join((component, 'mat'))))
            print('  -------------------')
            print(f'  {component}:')
            print('  MAE =', np.mean(res), 'nT')
            print('  RMSE =', np.sqrt(np.mean(res**2)), 'nT')
            print('  Max Error =', np.amax(res), 'nT')
            print('  Min Error =', np.amin(res), 'nT')

        self.assertIsNone(np.testing.assert_allclose(
            B_radius, B_radius_mat))
        self.assertIsNone(np.testing.assert_allclose(
            B_theta, B_theta_mat))
        self.assertIsNone(np.testing.assert_allclose(
            B_phi, B_phi_mat))

    def test_synth_sm_field(self):

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_synth_sm_field')

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        N = int(1000)

        time = np.linspace(-200, 6000, num=N)
        radius = R_REF
        theta = np.linspace(1, 179, num=N)
        phi = np.linspace(-180, 179, num=N)

        B_radius = np.zeros(time.shape)
        B_theta = np.zeros(time.shape)
        B_phi = np.zeros(time.shape)

        for source in ['internal', 'external']:
            B_radius_new, B_theta_new, B_phi_new = model.synth_values_sm(
                time, radius, theta, phi, source=source)

            B_radius += B_radius_new
            B_theta += B_theta_new
            B_phi += B_phi_new

        B_radius_mat = np.ravel(test['B_radius'])
        B_theta_mat = np.ravel(test['B_theta'])
        B_phi_mat = np.ravel(test['B_phi'])

        for component in ['B_radius', 'B_theta', 'B_phi']:
            res = np.abs(eval(component) - eval('_'.join((component, 'mat'))))
            print('  -------------------')
            print(f'  {component}:')
            print('  MAE =', np.mean(res), 'nT')
            print('  RMSE =', np.sqrt(np.mean(res**2)), 'nT')
            print('  Max Error =', np.amax(res), 'nT')
            print('  Min Error =', np.amin(res), 'nT')

        np.testing.assert_allclose(
            B_radius, B_radius_mat, rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(B_theta, B_theta_mat, rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(B_phi, B_phi_mat, rtol=1e-2, atol=1e-2)

    def test_synth_gsm_field(self):

        # load matfile
        test = load_matfile(MATFILE_PATH, 'test_synth_gsm_field')

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        N = int(1000)

        time = np.linspace(-1000, 6000, num=N)
        radius = R_REF
        theta = np.linspace(1, 179, num=N)
        phi = np.linspace(-180, 179, num=N)

        B_radius = np.zeros(time.shape)
        B_theta = np.zeros(time.shape)
        B_phi = np.zeros(time.shape)

        for source in ['internal', 'external']:
            B_radius_new, B_theta_new, B_phi_new = model.synth_values_gsm(
                time, radius, theta, phi, source=source)

            B_radius += B_radius_new
            B_theta += B_theta_new
            B_phi += B_phi_new

        B_radius_mat = np.ravel(test['B_radius'])
        B_theta_mat = np.ravel(test['B_theta'])
        B_phi_mat = np.ravel(test['B_phi'])

        for component in ['B_radius', 'B_theta', 'B_phi']:
            res = np.abs(eval(component) - eval('_'.join((component, 'mat'))))
            print('  -------------------')
            print(f'  {component}:')
            print('  MAE =', np.mean(res), 'nT')
            print('  RMSE =', np.sqrt(np.mean(res**2)), 'nT')
            print('  Max Error =', np.amax(res), 'nT')
            print('  Min Error =', np.amin(res), 'nT')

        np.testing.assert_allclose(
            B_radius, B_radius_mat, rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(
            B_theta, B_theta_mat, rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(
            B_phi, B_phi_mat, rtol=1e-2, atol=1e-2)

    def test_rotate_gsm_vector(self):

        time = 20
        nmax = 1
        kmax = 1
        radius = R_REF
        theta_geo = 90/6
        phi_geo = 90/8
        g_gsm = np.array([1, 2, -1])

        base_1, base_2, base_3 = cp.coordinate_utils.basevectors_gsm(time)
        matrix = cp.coordinate_utils.rotate_gauss(nmax, kmax, base_1,
                                                  base_2, base_3)

        g_geo = np.matmul(matrix, g_gsm)

        B_geo_1 = cp.model_utils.synth_values(g_geo, radius,
                                              theta_geo, phi_geo)
        B_geo_1 = np.array(B_geo_1)

        theta_gsm, phi_gsm, R = cp.coordinate_utils.matrix_geo_to_base(
            theta_geo, phi_geo, base_1, base_2, base_3)

        B_gsm = cp.model_utils.synth_values(g_gsm, radius, theta_gsm, phi_gsm)
        B_gsm = np.array(B_gsm)

        B_geo_2 = np.matmul(R.transpose(), B_gsm)

        np.testing.assert_allclose(B_geo_1, B_geo_2, rtol=1e-5)

    def test_plot_maps(self):

        model = cp.load_CHAOS_matfile(CHAOS_PATH)

        model.plot_maps_tdep(0., 6371.2)
        plt.close('all')

        model.plot_maps_static(6371.2, nmax=min(50, model.model_static.nmax))
        plt.close('all')


def profiler_complete_forward(n_data=300):
    """
    Example:

    .. code-block:: python

        %load_ext line_profiler

        import sys
        sys.path.insert(0, <tests directory path>)
        from test_chaosmagpy import profiler_complete_forward as profiler
        import chaosmagpy as cp

        %lprun -f cp.coordinate_utils.synth_rotate_gauss profiler(n_data=300)

    """

    t_start = -200.0
    t_end = 6000.0
    time = np.linspace(t_start, t_end, num=n_data)
    radius = R_REF * np.ones(time.shape)
    theta = np.linspace(1, 179, num=n_data)
    phi = np.linspace(-180, 179, num=n_data)

    model = cp.load_CHAOS_matfile(CHAOS_PATH)

    # B_radius, B_theta, B_phi = model(time, radius, theta, phi)
    B_radius, B_theta, B_phi = model(time, radius, theta, phi)
    print('Ran "profiler_complete_forward"')


if __name__ == '__main__':
    main()
